#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pytest_lazy_fixtures import lf

import numpy as np
from scipy.stats import unitary_group

import piquasso as pq

from jax import grad

from piquasso.decompositions.clements import (
    clements,
    inverse_clements,
    get_weights_from_decomposition,
    get_decomposition_from_weights,
    get_weights_from_interferometer,
    get_interferometer_from_weights,
    instructions_from_decomposition,
)


pytestmark = pytest.mark.monkey


@pytest.fixture
def dummy_unitary():
    def func(d):
        return np.array(unitary_group.rvs(d))

    return func


def test_clements_decomposition_using_piquasso_SamplingSimulator(dummy_unitary):
    d = 3
    U = dummy_unitary(d)

    decomposition = clements(U, connector=pq.NumpyConnector())

    with pq.Program() as program_with_interferometer:
        pq.Q() | pq.StateVector(tuple([1] * d))

        pq.Q() | pq.Interferometer(matrix=U)

    with pq.Program() as program_with_decomposition:
        pq.Q() | pq.StateVector(tuple([1] * d))

        for operation in decomposition.beamsplitters:
            pq.Q(operation.modes[0]) | pq.Phaseshifter(phi=operation.params[1])
            pq.Q(*operation.modes) | pq.Beamsplitter(operation.params[0], 0.0)

        for operation in decomposition.phaseshifters:
            pq.Q(operation.mode) | pq.Phaseshifter(operation.phi)

    simulator = pq.SamplingSimulator(d=d)

    state_with_interferometer = simulator.execute(program_with_interferometer).state
    state_with_decomposition = simulator.execute(program_with_decomposition).state

    assert state_with_interferometer == state_with_decomposition


def test_clements_decomposition_using_piquasso_PureFockSimulator(dummy_unitary):
    d = 4
    U = dummy_unitary(d)

    decomposition = clements(U, connector=pq.NumpyConnector())

    occupation_numbers = (1, 1, 0, 0)

    with pq.Program() as program_with_interferometer:
        pq.Q() | pq.StateVector(occupation_numbers)

        pq.Q() | pq.Interferometer(matrix=U)

    with pq.Program() as program_with_decomposition:
        pq.Q() | pq.StateVector(occupation_numbers)

        for operation in decomposition.beamsplitters:
            pq.Q(operation.modes[0]) | pq.Phaseshifter(phi=operation.params[1])
            pq.Q(*operation.modes) | pq.Beamsplitter(operation.params[0], 0.0)

        for operation in decomposition.phaseshifters:
            pq.Q(operation.mode) | pq.Phaseshifter(operation.phi)

    simulator = pq.PureFockSimulator(
        d=d, config=pq.Config(cutoff=sum(occupation_numbers) + 1)
    )

    state_with_interferometer = simulator.execute(program_with_interferometer).state
    state_with_decomposition = simulator.execute(program_with_decomposition).state

    assert state_with_interferometer == state_with_decomposition


def test_clements_decomposition_using_piquasso_FockSimulator(dummy_unitary):
    d = 4
    U = dummy_unitary(d)

    decomposition = clements(U, connector=pq.NumpyConnector())

    with pq.Program() as program_with_interferometer:
        pq.Q() | pq.DensityMatrix((1, 0, 1, 0), (1, 0, 1, 0))

        pq.Q() | pq.Interferometer(matrix=U)

    with pq.Program() as program_with_decomposition:
        pq.Q() | pq.DensityMatrix((1, 0, 1, 0), (1, 0, 1, 0))

        for operation in decomposition.beamsplitters:
            pq.Q(operation.modes[0]) | pq.Phaseshifter(phi=operation.params[1])
            pq.Q(*operation.modes) | pq.Beamsplitter(operation.params[0], 0.0)

        for operation in decomposition.phaseshifters:
            pq.Q(operation.mode) | pq.Phaseshifter(operation.phi)

    simulator = pq.FockSimulator(d=d, config=pq.Config(cutoff=3))

    state_with_interferometer = simulator.execute(program_with_interferometer).state
    state_with_decomposition = simulator.execute(program_with_decomposition).state

    assert state_with_interferometer == state_with_decomposition


def test_clements_decomposition_using_piquasso_GaussianSimulator(dummy_unitary):
    d = 3
    U = dummy_unitary(d)

    decomposition = clements(U, connector=pq.NumpyConnector())

    squeezings = [0.1, 0.2, 0.3]

    with pq.Program() as program_with_interferometer:
        for mode, r in enumerate(squeezings):
            pq.Q(mode) | pq.Squeezing(r)

        pq.Q() | pq.Interferometer(matrix=U)

    with pq.Program() as program_with_decomposition:
        for mode, r in enumerate(squeezings):
            pq.Q(mode) | pq.Squeezing(r)

        for operation in decomposition.beamsplitters:
            pq.Q(operation.modes[0]) | pq.Phaseshifter(phi=operation.params[1])
            pq.Q(*operation.modes) | pq.Beamsplitter(operation.params[0], 0.0)

        for operation in decomposition.phaseshifters:
            pq.Q(operation.mode) | pq.Phaseshifter(operation.phi)

    simulator = pq.GaussianSimulator(d=d)

    state_with_interferometer = simulator.execute(program_with_interferometer).state
    state_with_decomposition = simulator.execute(program_with_decomposition).state

    assert state_with_interferometer == state_with_decomposition


def test_instructions_from_decomposition_using_piquasso_GaussianSimulator(
    dummy_unitary,
):
    d = 3
    U = dummy_unitary(d)

    decomposition = clements(U, connector=pq.NumpyConnector())

    squeezings = [0.1, 0.2, 0.3]

    with pq.Program() as program_with_interferometer:
        for mode, r in enumerate(squeezings):
            pq.Q(mode) | pq.Squeezing(r)

        pq.Q() | pq.Interferometer(matrix=U)

    program_with_decomposition = pq.Program(
        instructions=[
            pq.Squeezing(r).on_modes(mode) for mode, r in enumerate(squeezings)
        ]
        + instructions_from_decomposition(decomposition)
    )

    simulator = pq.GaussianSimulator(d=d)

    state_with_interferometer = simulator.execute(program_with_interferometer).state
    state_with_decomposition = simulator.execute(program_with_decomposition).state

    assert state_with_interferometer == state_with_decomposition


def test_instructions_from_decomposition_using_piquasso_GaussianSimulator_with_context(
    dummy_unitary,
):
    d = 3
    U = dummy_unitary(d)

    decomposition = clements(U, connector=pq.NumpyConnector())

    squeezings = [0.1, 0.2, 0.3]

    with pq.Program() as program_with_interferometer:
        for mode, r in enumerate(squeezings):
            pq.Q(mode) | pq.Squeezing(r)

        pq.Q() | pq.Interferometer(matrix=U)

    with pq.Program() as program_with_decomposition:
        for mode, r in enumerate(squeezings):
            pq.Q(mode) | pq.Squeezing(r)

        program_with_decomposition.instructions.extend(
            instructions_from_decomposition(decomposition)
        )

    simulator = pq.GaussianSimulator(d=d)

    state_with_interferometer = simulator.execute(program_with_interferometer).state
    state_with_decomposition = simulator.execute(program_with_decomposition).state

    assert state_with_interferometer == state_with_decomposition


@pytest.mark.monkey
def test_instructions_from_decomposition_using_piquasso_GaussianSimulator_random(
    dummy_unitary, generate_gaussian_transform
):
    d = 3
    U = dummy_unitary(d)

    decomposition = clements(U, connector=pq.NumpyConnector())

    gaussian_transform = generate_gaussian_transform(d)

    with pq.Program() as program_with_interferometer:
        pq.Q() | gaussian_transform
        pq.Q() | pq.Interferometer(matrix=U)

    with pq.Program() as program_with_decomposition:
        pq.Q() | gaussian_transform
        program_with_decomposition.instructions.extend(
            instructions_from_decomposition(decomposition)
        )

    simulator = pq.GaussianSimulator(d=d)

    state_with_interferometer = simulator.execute(program_with_interferometer).state
    state_with_decomposition = simulator.execute(program_with_decomposition).state

    assert state_with_interferometer == state_with_decomposition


@pytest.mark.parametrize(
    "connector", (pq.NumpyConnector(), lf("tensorflow_connector"), pq.JaxConnector())
)
def test_clements_decomposition_roundtrip(connector, dummy_unitary):
    d = 5
    U = dummy_unitary(d)

    decomposition = clements(U, connector)

    new_U = inverse_clements(decomposition, connector, dtype=U.dtype)

    assert np.allclose(U, new_U)


@pytest.mark.parametrize(
    "connector", (pq.NumpyConnector(), lf("tensorflow_connector"), pq.JaxConnector())
)
def test_clements_decomposition_roundtrip_with_weights(connector, dummy_unitary):
    d = 6
    U = dummy_unitary(d)

    decomposition = clements(U, connector)

    weights = get_weights_from_decomposition(decomposition, d, connector)

    new_decomposition = get_decomposition_from_weights(weights, d, connector)

    assert decomposition == new_decomposition

    new_U = inverse_clements(new_decomposition, connector, dtype=U.dtype)

    assert np.allclose(U, new_U)


@pytest.mark.parametrize(
    "connector", (pq.NumpyConnector(), lf("tensorflow_connector"), pq.JaxConnector())
)
def test_weigths_to_interferometer_roundtrip(connector, dummy_unitary):
    d = 6
    U = connector.np.array(dummy_unitary(d))

    weights = get_weights_from_interferometer(U, connector)

    new_U = get_interferometer_from_weights(weights, d, connector, dtype=U.dtype)

    assert np.allclose(U, new_U)


def test_clements_decomposition_with_TensorflowConnector(dummy_unitary, tf):
    U = tf.Variable(dummy_unitary(2))

    with tf.GradientTape() as tape:
        decomposition = clements(U, connector=pq.TensorflowConnector())

    assert len(decomposition.beamsplitters) == 1
    assert len(decomposition.phaseshifters) == 2

    first_beamsplitter_theta = decomposition.beamsplitters[0].params[0]

    assert np.isclose(first_beamsplitter_theta, np.arctan(np.abs(U[1, 0] / U[0, 0])))

    gradient = tape.gradient(first_beamsplitter_theta, U)

    assert np.allclose(
        gradient,
        np.array(
            [
                [-np.abs(U[1, 0]) * U[0, 0] / np.abs(U[0, 0]), 0.0],
                [np.abs(U[0, 0]) * U[1, 0] / np.abs(U[1, 0]), 0.0],
            ]
        ),
    )


def test_clements_decomposition_with_JaxConnector(dummy_unitary):
    def get_first_beamsplitter_theta(U):
        decomposition = clements(U, connector=pq.JaxConnector())

        assert len(decomposition.beamsplitters) == 1
        assert len(decomposition.phaseshifters) == 2

        first_beamsplitter_theta = decomposition.beamsplitters[0].params[0]

        return first_beamsplitter_theta

    U = dummy_unitary(2)

    first_beamsplitter_theta = get_first_beamsplitter_theta(U)

    assert np.isclose(first_beamsplitter_theta, np.arctan(np.abs(U[1, 0] / U[0, 0])))

    grad_first_beamsplitter_theta = grad(get_first_beamsplitter_theta)

    gradient = grad_first_beamsplitter_theta(U)

    # NOTE: There is a difference between the differential with respect to a complex
    # variable between Tensorflow and JAX, namely, that the gradients in the two
    # frameworks are conjugate to each other.
    assert np.allclose(
        gradient,
        np.conj(
            np.array(
                [
                    [-np.abs(U[1, 0]) * U[0, 0] / np.abs(U[0, 0]), 0.0],
                    [np.abs(U[0, 0]) * U[1, 0] / np.abs(U[1, 0]), 0.0],
                ]
            )
        ),
    )


def test_clements_decomposition_with_TensorflowConnector_from_weights(
    dummy_unitary, tf
):
    d = 2
    U = dummy_unitary(d)

    weights = tf.Variable(get_weights_from_interferometer(U, pq.NumpyConnector()))

    with tf.GradientTape() as tape:
        decomposition = get_decomposition_from_weights(
            weights, d=d, connector=pq.TensorflowConnector()
        )

    assert len(decomposition.beamsplitters) == 1
    assert len(decomposition.phaseshifters) == 2

    first_beamsplitter_theta = decomposition.beamsplitters[0].params[0]

    assert np.isclose(first_beamsplitter_theta, np.arctan(np.abs(U[1, 0] / U[0, 0])))

    gradient = tape.gradient(first_beamsplitter_theta, weights)

    assert np.allclose(gradient, [1.0, 0.0, 0.0, 0.0])


def test_clements_decomposition_with_JaxConnector_from_weights(dummy_unitary):
    d = 2
    U = dummy_unitary(d)

    weights = get_weights_from_interferometer(U, pq.NumpyConnector())

    connector = pq.JaxConnector()

    def get_first_beamsplitter_theta(weights):
        decomposition = get_decomposition_from_weights(
            weights, d=d, connector=connector
        )

        first_beamsplitter_theta = decomposition.beamsplitters[0].params[0]

        return first_beamsplitter_theta

    first_beamsplitter_theta = get_first_beamsplitter_theta(weights)

    assert np.isclose(first_beamsplitter_theta, np.arctan(np.abs(U[1, 0] / U[0, 0])))

    grad_first_beamsplitter_theta = grad(get_first_beamsplitter_theta)

    gradient = grad_first_beamsplitter_theta(weights)

    assert np.allclose(
        gradient,
        [1.0, 0.0, 0.0, 0.0],
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An implementation for the Clements decomposition."""

from typing import List, Tuple, TYPE_CHECKING

from dataclasses import dataclass

from piquasso.api.connector import BaseConnector

from piquasso.instructions.gates import Phaseshifter, Beamsplitter

from piquasso._math.indices import get_operator_index

from piquasso._simulators.connectors import NumpyConnector

if TYPE_CHECKING:
    import numpy as np


@dataclass
class BS:
    r"""
    Beamsplitter gate, implemented as described in
    `arXiv:1603.08788 <https://arxiv.org/abs/1603.08788>`_.

    The single-particle unitary matrix corresponding to the beamsplitter is

    .. math::
        BS(\theta, \phi) = \begin{bmatrix}
            e^{i \phi} \cos \theta & - \sin \theta \\
            e^{i \phi} \sin \theta & \cos \theta
        \end{bmatrix}.

    Note:
        This means a different beamsplitter gate as defined by
        :class:`~piquasso.instructions.gates.Beamsplitter`.
    """

    modes: Tuple[int, int]
    params: Tuple["np.float64", "np.float64"]


@dataclass
class PS:
    """
    Phaseshifter gate, corresponding to
    :class:`~piquasso.instructions.gates.Phaseshifter`.
    """

    mode: int
    phi: "np.float64"


@dataclass
class Decomposition:
    """
    The data stucture which holds the decomposed angles from the Clements decomposition.

    Example usage::

        decomposition = clements(U, connector=pq.NumpyConnector())

        with pq.Program() as program:
            ...

            for operation in decomposition.beamsplitters:
                pq.Q(operation.modes[0]) | pq.Phaseshifter(phi=operation.params[1])
                pq.Q(*operation.modes) | pq.Beamsplitter(operation.params[0], 0.0)

            for operation in decomposition.phaseshifters:
                pq.Q(operation.mode) | pq.Phaseshifter(operation.phi)
    """

    beamsplitters: List[BS]
    phaseshifters: List[PS]


def instructions_from_decomposition(decomposition):
    """Helper function for using the Clements decomposition.

    This function creates the list of :class:`~piquasso.instructions.gates.Beamsplitter`
    and :class:`~piquasso.instructions.gates.Phaseshifter` instructions to be applied to
    the program, which is equivalent to the decomposed interferometer.

    Example usage::

        decomposition = clements(U, connector=pq.NumpyConnector())

        with pq.Program() as program:
            ...

            program.instructions.extend(instructions_from_decomposition(decomposition))

    Or, one can use it as::

        decomposition = clements(U, connector=pq.NumpyConnector())

        program_with_decomposition = pq.Program(
            instructions=[...] + instructions_from_decomposition(decomposition)
        )

    Args:
        decomposition (Decomposition): Decomposition created by :func:`clements`.

    Returns:
        list: List of beamsplitter and phaseshifter gates to be applied.
    """
    instructions = []

    for bs in decomposition.beamsplitters:
        instructions.append(Phaseshifter(bs.params[1]).on_modes(bs.modes[0]))
        instructions.append(Beamsplitter(bs.params[0], 0.0).on_modes(*bs.modes))

    for ps in decomposition.phaseshifters:
        instructions.append(Phaseshifter(ps.phi).on_modes(ps.mode))

    return instructions


def inverse_clements(
    decomposition: Decomposition, connector: BaseConnector, dtype: "np.dtype"
) -> "np.ndarray":
    """Inverse of the Clements decomposition.

    Returns:
        The unitary matrix corresponding to the interferometer.
    """

    d = len(decomposition.phaseshifters)

    np = connector.np
    interferometer = np.identity(d, dtype=dtype)

    for beamsplitter in decomposition.beamsplitters:
        beamsplitter_matrix = _get_embedded_beamsplitter_matrix(
            beamsplitter, d, connector, dtype=dtype
        )

        interferometer = beamsplitter_matrix @ interferometer

    phis = np.empty(d, dtype=interferometer.dtype)

    for phaseshifter in decomposition.phaseshifters:
        phis = connector.assign(phis, phaseshifter.mode, phaseshifter.phi)

    interferometer = np.diag(np.exp(1j * phis)) @ interferometer

    return interferometer


def clements(U: "np.ndarray", connector: BaseConnector) -> Decomposition:
    """
    Decomposes the specified unitary matrix by application of beamsplitters
    prescribed by the decomposition.

    Args:
        U (numpy.ndarray): The unitary matrix to be decomposed.

    Returns:
        The Clements decomposition. See :class:`Decomposition`.
    """

    first_beamsplitters = []
    last_beamsplitters = []

    np = connector.np

    d = U.shape[0]

    for column in reversed(range(0, d - 1)):
        if column % 2 == 0:
            operations, U = _apply_direct_beamsplitters(column, U, connector)
            last_beamsplitters.extend(operations)
        else:
            operations, U = _apply_inverse_beamsplitters(column, U, connector)
            first_beamsplitters.extend(operations)

    middle_phasshifters = [
        PS(mode=mode, phi=np.angle(diagonal))
        for mode, diagonal in enumerate(np.diag(U))
    ]

    last_beamsplitters = list(reversed(last_beamsplitters))

    commuted_beamsplitters, phasehifters = _commute(
        middle_phasshifters, last_beamsplitters, connector
    )

    return Decomposition(
        beamsplitters=first_beamsplitters + commuted_beamsplitters,
        phaseshifters=phasehifters,
    )


def _commute(
    middle_phaseshifters: List[PS],
    last_beamsplitters: List[BS],
    connector: BaseConnector,
) -> Tuple:
    r"""
    Rewrites the decomposition so that the phaseshifters are at the end of the circuit.

    One can commute an inverse beamsplitter and a layer of phaseshifters to another
    layer of phaseshifters and a different beamsplitter, as follows:

    .. math::
        BS(\theta, \phi)^{-1} D = D' BS(\theta', \phi')

    where :math:`D, D'` represent phaseshifters parametrized by
    :math:`\varphi_1, \varphi_2` and :math:`\varphi_1', \varphi_2'`, respectively.
    """

    commuted_beamsplitters = []

    for bs in last_beamsplitters:
        modes = bs.modes

        bs_theta_p, bs_phi_p, phi1_p, phi2_p = _get_commute_angles(
            bs.params[0],
            bs.params[1],
            middle_phaseshifters[modes[0]].phi,
            middle_phaseshifters[modes[1]].phi,
            connector,
        )

        bs2 = BS(modes=modes, params=(bs_theta_p, bs_phi_p))

        commuted_beamsplitters.append(bs2)
        middle_phaseshifters[modes[0]].phi = phi1_p
        middle_phaseshifters[modes[1]].phi = phi2_p

    return commuted_beamsplitters, middle_phaseshifters


def _get_commute_angles(bs_theta, bs_phi, phi1, phi2, connector):
    r"""Calculates the commutation angles.

    More concretely, this method calculates
    :math:`\theta', \phi', \varphi_1', \varphi_2'` from the relation

    .. math::
        BS(\theta, \phi)^{-1} D = D' BS(\theta', \phi')

    where :math:`D, D'` represent phaseshifters parametrized by
    :math:`\varphi_1, \varphi_2` and :math:`\varphi_1', \varphi_2'`, respectively.
    """

    np = connector.np

    bs_theta_p = bs_theta
    phi2_p = phi2

    bs_phi_p = np.mod(phi1 - phi2 + np.pi, 2 * np.pi)

    phi1_p = np.mod(phi2 - bs_phi + np.pi, 2 * np.pi)

    return bs_theta_p, bs_phi_p, phi1_p, phi2_p


def _apply_direct_beamsplitters(
    column: int, U: "np.ndarray", connector: BaseConnector
) -> tuple:
    """
    Calculates the direct beamsplitters for a given column `column`, and
    applies it to `U`.

    Args:
        column (int): The current column.
    """

    operations = []

    d = U.shape[0]

    dtype = U.dtype

    for j in range(d - 1 - column):
        modes = (column + j, column + j + 1)

        matrix_element_to_eliminate = U[modes[0], j]
        matrix_element_above = -U[modes[1], j]

        angles = _get_angles(
            matrix_element_to_eliminate, matrix_element_above, connector
        )

        operation = BS(modes=modes, params=angles)

        matrix = _get_embedded_beamsplitter_matrix(operation, d, connector, dtype)

        U = matrix @ U

        operations.append(operation)

    return operations, U


def _apply_inverse_beamsplitters(
    column: int, U: "np.ndarray", connector: BaseConnector
) -> tuple:
    """
    Calculates the inverse beamsplitters for a given column `column`, and
    applies it to `U`.

    Args:
        column (int): The current column.
    """

    operations = []

    d = U.shape[0]

    dtype = U.dtype

    for j in reversed(range(d - 1 - column)):
        modes = (j, j + 1)

        i = column + j + 1

        matrix_element_to_eliminate = U[i, modes[1]]
        matrix_element_to_left = U[i, modes[0]]

        angles = _get_angles(
            matrix_element_to_eliminate, matrix_element_to_left, connector
        )

        operation = BS(modes=modes, params=angles)

        beamsplitter = connector.np.conj(
            _get_embedded_beamsplitter_matrix(operation, d, connector, dtype)
        ).T

        U = U @ beamsplitter

        operations.append(operation)

    return operations, U


def _get_angles(matrix_element_to_eliminate, other_matrix_element, connector):
    np = connector.np

    if np.isclose(matrix_element_to_eliminate, 0.0):
        return np.pi / 2, 0.0

    r = other_matrix_element / matrix_element_to_eliminate
    theta = np.arctan(np.abs(r))
    phi = np.angle(r)

    return theta, phi


def _get_embedded_beamsplitter_matrix(
    operation: BS, d: int, connector: BaseConnector, dtype: "np.dtype"
) -> "np.ndarray":
    np = connector.np

    theta, phi = operation.params
    i, j = operation.modes

    c = np.cos(theta).astype(dtype)
    s = np.sin(theta).astype(dtype)

    matrix = np.array(
        [
            [np.exp(1j * phi) * c, -s],
            [np.exp(1j * phi) * s, c],
        ],
        dtype=dtype,
    )

    return connector.embed_in_identity(matrix, get_operator_index((i, j)), d)


def get_weights_from_decomposition(
    decomposition: Decomposition, d: int, connector: BaseConnector
) -> "np.ndarray":
    """Concatenates the weight vector from the angles in the Clements decomposition.

    Returns:
        The Clements decomposition. See :class:`Decomposition`.
    """
    np = connector.np

    dtype = decomposition.phaseshifters[0].phi.dtype
    weights = np.empty(d**2, dtype=dtype)

    index = 0
    for beamsplitter in decomposition.beamsplitters:
        weights = connector.assign(weights, index, beamsplitter.params[0])
        index += 1
        weights = connector.assign(weights, index, beamsplitter.params[1])
        index += 1

    for phaseshifter in decomposition.phaseshifters:
        weights = connector.assign(weights, index, phaseshifter.phi)
        index += 1

    return weights


def get_decomposition_from_weights(
    weights: "np.ndarray", d: int, connector: BaseConnector
) -> Decomposition:
    """Puts the data in the weight vector into a Clements decompositon.

    Returns:
        The Clements decomposition. See :class:`Decomposition`.
    """

    fallback_np = connector.fallback_np

    # NOTE: This is tricky: the ordering in the Clements decomposition is not unique,
    # since beamsplitters acting on different modes may commute, and the ordering comes
    # out very ugly after all the Givens rotations. Therefore, it is easier to just
    # create a trivial decomposition, and fill it with the required values (for now).
    decomposition = clements(fallback_np.identity(d), connector=NumpyConnector())

    index = 0

    for beamsplitter in decomposition.beamsplitters:
        beamsplitter.params = (weights[index], weights[index + 1])
        index += 2

    for phaseshifter in decomposition.phaseshifters:
        phaseshifter.phi = weights[index]
        index += 1

    return decomposition


def get_weights_from_interferometer(
    U: "np.ndarray", connector: BaseConnector
) -> "np.ndarray":
    """Creates a vector of weights from the Clements angles."""
    decomposition = clements(U, connector)

    return get_weights_from_decomposition(decomposition, U.shape[0], connector)


def get_interferometer_from_weights(
    weights: "np.ndarray",
    d: int,
    connector: BaseConnector,
    dtype: "np.dtype",
) -> "np.ndarray":
    """Returns the interferometer matrix corresponding to the specified weights.

    It is the inverse of :func:`get_weights_from_interferometer`.
    """
    decomposition = get_decomposition_from_weights(weights, d, connector)

    return inverse_clements(decomposition, connector, dtype)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time

import random

import matplotlib.pyplot as plt

import numpy as np

import perceval as pcvl
from perceval.algorithm import Sampler

import piquasso as pq


N = 100  # number of samplings


# Copied from: https://perceval.quandela.net/docs/v0.11/notebooks/Boson_sampling.html


def Generating_Input(n, m):
    "This function randomly chooses an input with n photons in m modes."
    modes = sorted(random.sample(range(m), n))

    state = "|"
    for i in range(m):
        state = (
            state + "0" * (1 - (i in modes)) + "1" * (i in modes) + "," * (i < m - 1)
        )
    return pcvl.BasicState(state + ">")


mzi = (
    pcvl.BS()
    // (0, pcvl.PS(phi=pcvl.Parameter("φ_a")))
    // pcvl.BS()
    // (1, pcvl.PS(phi=pcvl.Parameter("φ_b")))
)


def get_perceval_samples(unitary):
    m = len(unitary)

    n = m // 2

    Linear_Circuit = pcvl.Circuit.decomposition(
        unitary,
        mzi,
        phase_shifter_fn=pcvl.PS,
        shape=pcvl.InterferometerShape.TRIANGLE,
    )

    QPU = pcvl.Processor("CliffordClifford2017", Linear_Circuit)

    input_state = Generating_Input(n, m)
    QPU.with_input(input_state)

    # Keep all outputs
    QPU.min_detected_photons_filter(0)

    sampler = Sampler(QPU)

    return sampler.samples(N)["results"]


def generate_random_fock_state(m, n):
    modes = sorted(random.sample(range(m), n))

    state_vector = []
    for i in range(m):
        if i in modes:
            state_vector.append(1)
        else:
            state_vector.append(0)

    return state_vector


def get_piquasso_samples(unitary):
    m = len(unitary)
    n = m // 2

    input_state = generate_random_fock_state(m=m, n=n)
    program = pq.Program(
        instructions=[
            pq.StateVector(input_state),
            pq.Interferometer(unitary),
            pq.ParticleNumberMeasurement(),
        ]
    )

    simulator = pq.SamplingSimulator(d=m)

    return simulator.execute(program, shots=N).samples


if __name__ == "__main__":

    # Warmup
    m = 2
    unitary = pcvl.Matrix.random_unitary(m)
    get_perceval_samples(unitary)
    get_piquasso_samples(np.array(unitary))
    ####

    x = []
    pv_times = []
    pq_times = []

    for m in range(2, 30):
        print("m=", m)
        x.append(m)
        unitary = pcvl.Matrix.random_unitary(m)

        start_time = time.time()
        get_perceval_samples(unitary)
        runtime = time.time() - start_time
        print("PV:", runtime)
        pv_times.append(runtime)

        start_time = time.time()
        get_piquasso_samples(np.array(unitary))
        runtime = time.time() - start_time
        print("PQ:", runtime)
        pq_times.append(runtime)

    plt.scatter(x, pv_times, label="Perceval")
    plt.scatter(x, pq_times, label="Piquasso")

    plt.yscale("log")

    plt.legend()
    plt.show()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from pytest_lazy_fixtures import lf


connectors = (lf("tensorflow_connector"), lf("tensorflow_connector_tf_function"))


@pytest.mark.parametrize("connector", connectors)
def test_batch_Beamsplitter_mean_position(connector, tf):
    theta = tf.Variable(np.pi / 3)

    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with tf.GradientTape() as tape:
        with pq.Program() as batch_program:
            pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

            pq.Q() | pq.Beamsplitter(theta=theta, phi=np.pi / 3)

        simulator = pq.PureFockSimulator(
            d=2, config=pq.Config(cutoff=5), connector=connector
        )

        batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    batch_gradient = tape.jacobian(batch_mean_positions, theta)

    with tf.GradientTape() as tape:
        with pq.Program() as first_program:
            pq.Q() | first_preparation

            pq.Q() | pq.Beamsplitter(theta=theta, phi=np.pi / 3)

        first_mean_position = simulator.execute(first_program).state.mean_position(0)

    first_gradient = tape.gradient(first_mean_position, theta)

    with tf.GradientTape() as tape:
        with pq.Program() as second_program:
            pq.Q() | second_preparation

            pq.Q() | pq.Beamsplitter(theta=theta, phi=np.pi / 3)

        second_mean_position = simulator.execute(second_program).state.mean_position(0)

    second_gradient = tape.gradient(second_mean_position, theta)

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)

    assert np.allclose(batch_gradient[0], first_gradient)
    assert np.allclose(batch_gradient[1], second_gradient)


@pytest.mark.parametrize("connector", connectors)
def test_batch_Squeezing_mean_position(connector, tf):
    r = tf.Variable(0.1)

    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with tf.GradientTape() as tape:
        with pq.Program() as batch_program:
            pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

            pq.Q(0) | pq.Squeezing(r=r)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        simulator = pq.PureFockSimulator(
            d=2, config=pq.Config(cutoff=5), connector=connector
        )

        batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    batch_gradient = tape.jacobian(batch_mean_positions, r)

    with tf.GradientTape() as tape:
        with pq.Program() as first_program:
            pq.Q() | first_preparation

            pq.Q(0) | pq.Squeezing(r=r)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        first_mean_position = simulator.execute(first_program).state.mean_position(0)

    first_gradient = tape.gradient(first_mean_position, r)

    with tf.GradientTape() as tape:
        with pq.Program() as second_program:
            pq.Q() | second_preparation

            pq.Q(0) | pq.Squeezing(r=r)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        second_mean_position = simulator.execute(second_program).state.mean_position(0)

    second_gradient = tape.gradient(second_mean_position, r)

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)

    assert np.allclose(batch_gradient[0], first_gradient)
    assert np.allclose(batch_gradient[1], second_gradient)


@pytest.mark.parametrize("connector", connectors)
def test_batch_Displacement_mean_position(connector, tf):
    r = tf.Variable(0.1)

    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with tf.GradientTape() as tape:
        with pq.Program() as batch_program:
            pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

            pq.Q(0) | pq.Displacement(r=r)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        simulator = pq.PureFockSimulator(
            d=2, config=pq.Config(cutoff=5), connector=connector
        )

        batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    batch_gradient = tape.jacobian(batch_mean_positions, r)

    with tf.GradientTape() as tape:
        with pq.Program() as first_program:
            pq.Q() | first_preparation

            pq.Q(0) | pq.Displacement(r=r)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        first_mean_position = simulator.execute(first_program).state.mean_position(0)

    first_gradient = tape.gradient(first_mean_position, r)

    with tf.GradientTape() as tape:
        with pq.Program() as second_program:
            pq.Q() | second_preparation

            pq.Q(0) | pq.Displacement(r=r)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        second_mean_position = simulator.execute(second_program).state.mean_position(0)

    second_gradient = tape.gradient(second_mean_position, r)

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)

    assert np.allclose(batch_gradient[0], first_gradient)
    assert np.allclose(batch_gradient[1], second_gradient)


@pytest.mark.parametrize("connector", connectors)
def test_batch_Kerr_mean_position(connector, tf):
    xi = tf.Variable(0.1)

    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with tf.GradientTape() as tape:
        with pq.Program() as batch_program:
            pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

            pq.Q(0) | pq.Kerr(xi=xi)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        simulator = pq.PureFockSimulator(
            d=2, config=pq.Config(cutoff=5), connector=connector
        )

        batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    batch_gradient = tape.jacobian(batch_mean_positions, xi)

    with tf.GradientTape() as tape:
        with pq.Program() as first_program:
            pq.Q() | first_preparation

            pq.Q(0) | pq.Kerr(xi=xi)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        first_mean_position = simulator.execute(first_program).state.mean_position(0)

    first_gradient = tape.gradient(first_mean_position, xi)

    with tf.GradientTape() as tape:
        with pq.Program() as second_program:
            pq.Q() | second_preparation

            pq.Q(0) | pq.Kerr(xi=xi)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        second_mean_position = simulator.execute(second_program).state.mean_position(0)

    second_gradient = tape.gradient(second_mean_position, xi)

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)

    assert np.allclose(batch_gradient[0], first_gradient)
    assert np.allclose(batch_gradient[1], second_gradient)


@pytest.mark.parametrize("connector", connectors)
def test_batch_Phaseshifter_mean_position(connector, tf):
    phi = tf.Variable(np.pi / 5)

    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with tf.GradientTape() as tape:
        with pq.Program() as batch_program:
            pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

            pq.Q(0) | pq.Phaseshifter(phi=phi)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        simulator = pq.PureFockSimulator(
            d=2, config=pq.Config(cutoff=5), connector=connector
        )

        batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    batch_gradient = tape.jacobian(batch_mean_positions, phi)

    with tf.GradientTape() as tape:
        with pq.Program() as first_program:
            pq.Q() | first_preparation

            pq.Q(0) | pq.Phaseshifter(phi=phi)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        first_mean_position = simulator.execute(first_program).state.mean_position(0)

    first_gradient = tape.gradient(first_mean_position, phi)

    with tf.GradientTape() as tape:
        with pq.Program() as second_program:
            pq.Q() | second_preparation

            pq.Q(0) | pq.Phaseshifter(phi=phi)
            pq.Q() | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 3)

        second_mean_position = simulator.execute(second_program).state.mean_position(0)

    second_gradient = tape.gradient(second_mean_position, phi)

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)

    assert np.allclose(batch_gradient[0], first_gradient)
    assert np.allclose(batch_gradient[1], second_gradient)


@pytest.mark.parametrize("connector", connectors)
def test_batch_complex_circuit_mean_position(connector, tf):
    theta1 = tf.Variable(np.pi / 3)
    phi1 = tf.Variable(np.pi / 4)

    phi_phaseshift = tf.Variable(np.pi / 5)

    r1 = tf.Variable(0.1)
    r2 = tf.Variable(0.2)

    alpha1 = tf.Variable(1.0)
    alpha2 = tf.Variable(0.5)

    theta2 = tf.Variable(np.pi / 3)
    phi2 = tf.Variable(np.pi / 4)

    theta2 = tf.Variable(np.pi / 3)
    phi2 = tf.Variable(np.pi / 4)

    xi1 = tf.Variable(0.1)
    xi2 = tf.Variable(0.15)

    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with tf.GradientTape() as tape:
        with pq.Program() as batch_program:
            pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

            pq.Q() | pq.Beamsplitter(theta=theta1, phi=phi1)
            pq.Q(0) | pq.Phaseshifter(phi=phi_phaseshift)

            pq.Q(0) | pq.Squeezing(r=r1)
            pq.Q(1) | pq.Squeezing(r=r2)

            pq.Q(0) | pq.Displacement(r=alpha1)
            pq.Q(1) | pq.Displacement(r=alpha2)

            pq.Q() | pq.Beamsplitter(theta=theta2, phi=phi2)

            pq.Q(0) | pq.Kerr(xi=xi1)
            pq.Q(1) | pq.Kerr(xi=xi2)

        simulator = pq.PureFockSimulator(
            d=2, config=pq.Config(cutoff=5), connector=connector
        )

        batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    parameters = [theta1, phi1, phi_phaseshift, r1, r2, alpha1, alpha2, theta2, phi2]

    batch_gradient = np.array(tape.jacobian(batch_mean_positions, parameters))

    with tf.GradientTape() as tape:
        with pq.Program() as first_program:
            pq.Q() | first_preparation

            pq.Q() | pq.Beamsplitter(theta=theta1, phi=phi1)
            pq.Q(0) | pq.Phaseshifter(phi=phi_phaseshift)

            pq.Q(0) | pq.Squeezing(r=r1)
            pq.Q(1) | pq.Squeezing(r=r2)

            pq.Q(0) | pq.Displacement(r=alpha1)
            pq.Q(1) | pq.Displacement(r=alpha2)

            pq.Q() | pq.Beamsplitter(theta=theta2, phi=phi2)

            pq.Q(0) | pq.Kerr(xi=xi1)
            pq.Q(1) | pq.Kerr(xi=xi2)

        first_mean_position = simulator.execute(first_program).state.mean_position(0)

    first_gradient = np.array(tape.gradient(first_mean_position, parameters))

    with tf.GradientTape() as tape:
        with pq.Program() as second_program:
            pq.Q() | second_preparation

            pq.Q() | pq.Beamsplitter(theta=theta1, phi=phi1)
            pq.Q(0) | pq.Phaseshifter(phi=phi_phaseshift)

            pq.Q(0) | pq.Squeezing(r=r1)
            pq.Q(1) | pq.Squeezing(r=r2)

            pq.Q(0) | pq.Displacement(r=alpha1)
            pq.Q(1) | pq.Displacement(r=alpha2)

            pq.Q() | pq.Beamsplitter(theta=theta2, phi=phi2)

            pq.Q(0) | pq.Kerr(xi=xi1)
            pq.Q(1) | pq.Kerr(xi=xi2)

        second_mean_position = simulator.execute(second_program).state.mean_position(0)

    second_gradient = np.array(tape.gradient(second_mean_position, parameters))

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)

    assert np.allclose(batch_gradient[:, 0], first_gradient)
    assert np.allclose(batch_gradient[:, 1], second_gradient)


@pytest.mark.parametrize("connector", connectors)
def test_batch_complex_circuit_mean_position_with_batch_apply(connector, tf):
    theta1 = tf.Variable(np.pi / 3)
    phi1 = tf.Variable(np.pi / 4)

    phi_phaseshift = tf.Variable(np.pi / 5)

    r1 = tf.Variable(0.1)
    r2 = tf.Variable(0.2)

    alpha1 = tf.Variable(1.0)
    alpha2 = tf.Variable(0.5)

    theta2 = tf.Variable(np.pi / 3)
    phi2 = tf.Variable(np.pi / 4)

    theta2 = tf.Variable(np.pi / 3)
    phi2 = tf.Variable(np.pi / 4)

    xi1 = tf.Variable(0.1)
    xi2 = tf.Variable(0.15)

    xi1_intermediate = tf.Variable(0.3)
    xi2_intermediate = tf.Variable(0.4)

    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with tf.GradientTape() as tape:
        with pq.Program() as first_intermediate:
            pq.Q(0) | pq.Kerr(xi1_intermediate)

        with pq.Program() as second_intermediate:
            pq.Q(1) | pq.Kerr(xi2_intermediate)

        with pq.Program() as batch_program:
            pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

            pq.Q() | pq.Beamsplitter(theta=theta1, phi=phi1)
            pq.Q(0) | pq.Phaseshifter(phi=phi_phaseshift)

            pq.Q(0) | pq.Squeezing(r=r1)
            pq.Q(1) | pq.Squeezing(r=r2)

            pq.Q() | pq.BatchApply([first_intermediate, second_intermediate])

            pq.Q(0) | pq.Displacement(r=alpha1)
            pq.Q(1) | pq.Displacement(r=alpha2)

            pq.Q() | pq.Beamsplitter(theta=theta2, phi=phi2)

            pq.Q(0) | pq.Kerr(xi=xi1)
            pq.Q(1) | pq.Kerr(xi=xi2)

        simulator = pq.PureFockSimulator(
            d=2, config=pq.Config(cutoff=5), connector=connector
        )

        batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    parameters = [theta1, phi1, phi_phaseshift, r1, r2, alpha1, alpha2, theta2, phi2]

    batch_gradient = np.array(
        tape.jacobian(
            batch_mean_positions, parameters + [xi1_intermediate, xi2_intermediate]
        )
    )

    with tf.GradientTape() as tape:
        with pq.Program() as first_program:
            pq.Q() | first_preparation

            pq.Q() | pq.Beamsplitter(theta=theta1, phi=phi1)
            pq.Q(0) | pq.Phaseshifter(phi=phi_phaseshift)

            pq.Q(0) | pq.Squeezing(r=r1)
            pq.Q(1) | pq.Squeezing(r=r2)

            pq.Q(0) | pq.Kerr(xi1_intermediate)

            pq.Q(0) | pq.Displacement(r=alpha1)
            pq.Q(1) | pq.Displacement(r=alpha2)

            pq.Q() | pq.Beamsplitter(theta=theta2, phi=phi2)

            pq.Q(0) | pq.Kerr(xi=xi1)
            pq.Q(1) | pq.Kerr(xi=xi2)

        first_mean_position = simulator.execute(first_program).state.mean_position(0)

    first_gradient = np.array(
        tape.gradient(first_mean_position, parameters + [xi1_intermediate])
    )

    with tf.GradientTape() as tape:
        with pq.Program() as second_program:
            pq.Q() | second_preparation

            pq.Q() | pq.Beamsplitter(theta=theta1, phi=phi1)
            pq.Q(0) | pq.Phaseshifter(phi=phi_phaseshift)

            pq.Q(0) | pq.Squeezing(r=r1)
            pq.Q(1) | pq.Squeezing(r=r2)

            pq.Q(1) | pq.Kerr(xi2_intermediate)

            pq.Q(0) | pq.Displacement(r=alpha1)
            pq.Q(1) | pq.Displacement(r=alpha2)

            pq.Q() | pq.Beamsplitter(theta=theta2, phi=phi2)

            pq.Q(0) | pq.Kerr(xi=xi1)
            pq.Q(1) | pq.Kerr(xi=xi2)

        second_mean_position = simulator.execute(second_program).state.mean_position(0)

    second_gradient = np.array(
        tape.gradient(second_mean_position, parameters + [xi2_intermediate])
    )

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)

    assert np.allclose(
        batch_gradient[: len(parameters), 0], first_gradient[: len(parameters)]
    )
    assert np.isclose(batch_gradient[-2, 0], first_gradient[-1])
    assert np.isclose(batch_gradient[-1, 0], 0.0)

    assert np.allclose(
        batch_gradient[: len(parameters), 1], second_gradient[: len(parameters)]
    )
    assert np.isclose(batch_gradient[-1, 1], second_gradient[-1])
    assert np.isclose(batch_gradient[-2, 1], 0.0)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time

from scipy.stats import unitary_group
import matplotlib.pyplot as plt

import json

import numpy as np

import piquasso as pq


SHOTS = 100  # number of shots


def get_piquasso_samples(squeezings, unitary, use_dask):
    d = len(squeezings)

    with pq.Program() as program:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(squeezings[i])

        pq.Q() | pq.Interferometer(unitary)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(
        d=d, config=pq.Config(measurement_cutoff=5 + 1, use_dask=use_dask)
    )

    return simulator.execute(program, shots=SHOTS).samples


if __name__ == "__main__":
    FILENAME = f"gaussian_boson_sampling_{int(time.time())}.json"

    np.random.seed(123)

    # Warmup
    m = 2
    squeezings = np.arcsinh(np.ones(m))
    unitary = unitary_group.rvs(m)
    get_piquasso_samples(squeezings, unitary, use_dask=True)
    get_piquasso_samples(squeezings, unitary, use_dask=False)
    ####

    x = []
    pq_dask_times = []
    pq_times = []

    for m in range(25, 26):
        print("m=", m)
        x.append(m)

        squeezings = np.arcsinh(np.ones(m))
        unitary = unitary_group.rvs(m)

        start_time = time.time()
        samples = get_piquasso_samples(squeezings, unitary, use_dask=True)
        runtime = time.time() - start_time
        print("PQ (dask):", runtime)
        pq_dask_times.append(runtime)

        start_time = time.time()
        samples = get_piquasso_samples(squeezings, unitary, use_dask=False)
        runtime = time.time() - start_time
        print("PQ:", runtime)
        pq_times.append(runtime)

        with open(FILENAME, "w") as f:
            json.dump(
                dict(x=x, pq_times=pq_times, pq_dask_times=pq_dask_times), f, indent=4
            )

    plt.scatter(x, pq_dask_times, label="Piquasso (dask)")
    plt.scatter(x, pq_times, label="Piquasso")

    plt.yscale("log")

    plt.legend()
    plt.show()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Any

import os
import copy
import random
import numpy as np

from piquasso.core import _mixins


class Config(_mixins.CodeMixin):
    """The configuration for the simulation.

    :ivar cutoff: The Fock space cutoff. Defaults to `4`.
    :ivar dtype:
        The underlying datatype of the simulation. Possible values: `np.float32` and
        `np.float64`. Defaults to `np.float64`.
    :ivar measurement_cutoff:
        The maximum number of particles to be allowed for
        :class:`~piquasso.instructions.measurements.ParticleNumberMeasurement`
        using :class:`~piquasso._simulators.gaussian.simulator.GaussianSimulator`.
        Defaults to `5`.
    :ivar hbar: The value of the Planck constant. Defaults to `2.0`.
    :ivar seed_sequence: The seed for reproducability of sampling algorithms.
    :ivar use_torontonian:
        Uses torontonian for
        :class:`~piquasso.instructions.measurements.ThresholdMeasurement`. Defaults to
        `False`.
    :ivar cache_size:
        The maximum size of the cache for certain algorithms. Defaults to `2.0`.
    :ivar validate:
        Validates computations during simulation. Defaults to `True`. If set to `False`,
        it is not guaranteed that the calculations will be correct, and it is advised
        to only turn it off when necessary. Moreover, it is also not guaranteed that all
        validations are turned off by setting `validate=False` (e.g., specifying invalid
        modes will still yield an error).
    :ivar use_dask:
        Use dask for certain computations during the simulation, e.g., for Gaussian
        Boson Sampling. Defaults to `False`.
    """

    def __init__(
        self,
        *,
        cutoff: int = 4,
        dtype: type = np.float64,
        measurement_cutoff: int = 5,
        hbar: float = 2.0,
        seed_sequence: Optional[Any] = None,
        use_torontonian: bool = False,
        cache_size: int = 32,
        validate: bool = True,
        use_dask: bool = False,
    ):
        self._original_seed_sequence = seed_sequence
        self.seed_sequence = seed_sequence or int.from_bytes(
            os.urandom(8), byteorder="big"
        )
        self.cache_size = cache_size
        self.hbar = hbar
        self.use_torontonian = use_torontonian
        self.cutoff = cutoff
        self.measurement_cutoff = measurement_cutoff
        self.dtype = np.float64 if dtype is float else dtype
        self.validate = validate
        self.use_dask = use_dask

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Config):
            return False
        return (
            self._original_seed_sequence == other._original_seed_sequence
            and self.cache_size == other.cache_size
            and self.hbar == other.hbar
            and self.use_torontonian == other.use_torontonian
            and self.cutoff == other.cutoff
            and self.measurement_cutoff == other.measurement_cutoff
            and self.dtype == other.dtype
            and self.validate == other.validate
            and self.use_dask == other.use_dask
        )

    def _as_code(self) -> str:
        default_config = Config()
        non_default_params = dict()

        if self._original_seed_sequence != default_config._original_seed_sequence:
            non_default_params["seed_sequence"] = self._original_seed_sequence
        if self.cache_size != default_config.cache_size:
            non_default_params["cache_size"] = self.cache_size
        if self.hbar != default_config.hbar:
            non_default_params["hbar"] = self.hbar
        if self.use_torontonian != default_config.use_torontonian:
            non_default_params["use_torontonian"] = self.use_torontonian
        if self.cutoff != default_config.cutoff:
            non_default_params["cutoff"] = self.cutoff
        if self.measurement_cutoff != default_config.measurement_cutoff:
            non_default_params["measurement_cutoff"] = self.measurement_cutoff
        if self.dtype != default_config.dtype:
            non_default_params["dtype"] = "np." + self.dtype.__name__
        if self.validate != default_config.validate:
            non_default_params["validate"] = self.validate
        if self.use_dask != default_config.use_dask:
            non_default_params["use_dask"] = self.use_dask

        if len(non_default_params) == 0:
            return "pq.Config()"
        else:
            params_string = ", ".join(
                f"{key}={value}" for key, value in non_default_params.items()
            )
            return f"pq.Config({params_string})"

    @property
    def seed_sequence(self):
        """The seed sequence used to generate random numbers during the simulation."""

        return self._seed_sequence

    @seed_sequence.setter
    def seed_sequence(self, value: Any) -> None:
        self._seed_sequence = value
        self.rng = np.random.default_rng(self._seed_sequence)
        random.seed(self._seed_sequence)

    @property
    def complex_dtype(self):
        """Returns the complex precision depending on the dtype of the Config class"""

        if self.dtype is np.float64:
            return np.complex128

        return np.complex64

    def copy(self) -> "Config":
        """Returns an exact copy of this config object.

        Returns:
            Config: An exact copy of this config object.
        """

        config_copy = copy.deepcopy(self)

        # NOTE: We want to preserve the RNG, otherwise simulations may lead to repeated
        # samples if the user reuses the simulator.
        config_copy.rng = self.rng

        return config_copy

    def __repr__(self):
        return self._as_code()[3:]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq
import tensorflow as tf
import time
from scipy.stats import unitary_group


alpha = 0.01
r = 0.01
xi = 0.3
d = 2

interferometer = unitary_group.rvs(d)

alpha_ = tf.Variable(alpha, dtype=tf.float32)

with pq.Program() as program:
    pq.Q(all) | pq.Vacuum()

    pq.Q(all) | pq.Displacement(r=alpha_)
    pq.Q(all) | pq.Squeezing(r)
    pq.Q(all) | pq.Interferometer(interferometer)
    pq.Q(all) | pq.Kerr(xi)

simulator_fock = pq.PureFockSimulator(
    d=d, config=pq.Config(cutoff=d), connector=pq.TensorflowConnector()
)

with tf.GradientTape() as tape:
    state = simulator_fock.execute(program).state
    fock_probabilities = state.fock_probabilities


profiler_options = tf.profiler.experimental.ProfilerOptions(
    host_tracer_level=3, python_tracer_level=3, device_tracer_level=3
)

tf.profiler.experimental.start("logdir", options=profiler_options)

start_time = time.time()
gradient = tape.jacobian(fock_probabilities, [alpha_])
print("JACOBIAN CALCULATION TIME: ", time.time() - start_time)

tf.profiler.experimental.stop()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.api.instruction import Preparation, Gate, BatchInstruction


class BatchPrepare(Preparation, BatchInstruction):
    r"""Allows for batch processing of multiple states.

    NOTE: This feature is experimental.
    """

    def __init__(self, subprograms):
        super().__init__(params=dict(subprograms=subprograms))


class BatchApply(Gate, BatchInstruction):
    r"""Applies subprograms to each element in the batch.

    NOTE: This feature is experimental.
    """

    def __init__(self, subprograms):
        super().__init__(params=dict(subprograms=subprograms))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from jax import grad
import jax.numpy as jnp


@pytest.mark.monkey
@pytest.mark.parametrize("SimulatorClass", (pq.PureFockSimulator, pq.SamplingSimulator))
def test_Jax_gradient_equivalence_for_single_Beamsplitter(SimulatorClass):
    connector = pq.JaxConnector()

    def get_fidelity(theta):
        initial_state = [1, 1, 0]

        d = len(initial_state)
        cutoff = np.sum(initial_state) + 1

        with pq.Program() as program:
            pq.Q() | pq.StateVector(initial_state)

        with pq.Program() as rotated_program:
            pq.Q() | program

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta)

        simulator = SimulatorClass(
            d=d,
            connector=connector,
            config=pq.Config(cutoff=cutoff),
        )

        initial_state_vector = simulator.execute(program).state.state_vector
        rotated_state_vector = simulator.execute(rotated_program).state.state_vector

        return jnp.abs(jnp.conj(initial_state_vector) @ rotated_state_vector)

    angle = np.random.uniform() * 2 * np.pi

    expected_fidelity = np.abs(np.cos(2 * angle))

    fidelity = get_fidelity(angle)

    assert np.isclose(fidelity, expected_fidelity)

    grad_get_fidelity = grad(get_fidelity)
    fidelity_grad = grad_get_fidelity(angle)

    expected_fidelity_grad = -2 * np.sin(2 * angle) * np.sign(np.cos(2 * angle))

    assert np.isclose(fidelity_grad, expected_fidelity_grad)


@pytest.mark.parametrize("SimulatorClass", (pq.PureFockSimulator, pq.SamplingSimulator))
def test_Jax_gradient_equivalence_complex_scenario(SimulatorClass):
    connector = pq.JaxConnector()

    def get_fidelity(theta):
        initial_state = [1, 1, 0]

        d = len(initial_state)
        cutoff = np.sum(initial_state) + 1

        with pq.Program() as program:
            pq.Q() | pq.StateVector(initial_state)

        with pq.Program() as rotated_program:
            pq.Q() | program

            pq.Q(1, 2) | pq.Beamsplitter(theta=0.1, phi=0.42)

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta)

            pq.Q(1, 2) | pq.Beamsplitter(theta=1.5, phi=0.35)

        simulator = SimulatorClass(
            d=d,
            connector=connector,
            config=pq.Config(cutoff=cutoff),
        )

        initial_state_vector = simulator.execute(program).state.state_vector
        rotated_state_vector = simulator.execute(rotated_program).state.state_vector

        return jnp.abs(jnp.conj(initial_state_vector) @ rotated_state_vector)

    angle = np.pi / 7

    fidelity = get_fidelity(angle)

    expected_fidelity = 0.04604777

    assert np.isclose(fidelity, expected_fidelity)

    grad_get_fidelity = grad(get_fidelity)
    fidelity_grad = grad_get_fidelity(angle)

    expected_fidelity_grad = 0.06591828

    assert np.isclose(fidelity_grad, expected_fidelity_grad)


@pytest.mark.parametrize("SimulatorClass", (pq.PureFockSimulator, pq.SamplingSimulator))
def test_PostSelectPhotons_gradient(SimulatorClass):
    connector = pq.JaxConnector()
    state_vector = np.sqrt([0.2, 0.3, 0.5])

    def _calculate_loss(weights):
        np = connector.np

        with pq.Program() as preparation:
            pq.Q(all) | pq.StateVector([0, 1, 0]) * state_vector[0]
            pq.Q(all) | pq.StateVector([1, 1, 0]) * state_vector[1]
            pq.Q(all) | pq.StateVector([2, 1, 0]) * state_vector[2]

        phase_shifter_phis = weights[:3]
        thetas = weights[3:6]
        phis = weights[6:]
        with pq.Program() as interferometer:
            for i in range(3):
                pq.Q(i) | pq.Phaseshifter(phase_shifter_phis[i])

            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[0], phi=phis[0])
            pq.Q(0, 1) | pq.Beamsplitter(theta=thetas[1], phi=phis[1])
            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[2], phi=phis[2])

        with pq.Program() as program:
            pq.Q(all) | preparation

            pq.Q(all) | interferometer

            pq.Q(all) | pq.PostSelectPhotons(
                postselect_modes=(1, 2), photon_counts=(1, 0)
            )

        simulator = SimulatorClass(d=3, config=pq.Config(cutoff=4), connector=connector)

        state = simulator.execute(program).state

        norm = state.norm

        density_matrix = state.density_matrix[:3, :3] / norm

        expected_state = np.copy(state_vector)
        expected_state = connector.assign(expected_state, 2, -expected_state[2])

        loss = 1 - np.sqrt(
            np.real(np.conj(expected_state) @ density_matrix @ expected_state)
        )

        return loss

    weights = np.array(
        [np.pi, 0.0, 0.0, np.pi / 8, 65.5302 * 2 * np.pi / 360, -np.pi / 8, 0, 0, 0]
    )

    loss = _calculate_loss(weights=weights)

    loss_grad = grad(_calculate_loss)

    gradient = loss_grad(weights)

    assert np.isclose(loss, 0.0)

    assert np.allclose(gradient, 0.0, atol=1e-7)


@pytest.mark.parametrize("SimulatorClass", (pq.PureFockSimulator, pq.SamplingSimulator))
def test_ImperfectPostSelectPhotons_gradient(SimulatorClass):

    connector = pq.JaxConnector()

    state_vector = np.sqrt([0.2, 0.3, 0.5])

    def _calculate_loss(weights):
        np = connector.np

        with pq.Program() as preparation:
            pq.Q(all) | pq.StateVector([0, 1, 0]) * state_vector[0]
            pq.Q(all) | pq.StateVector([1, 1, 0]) * state_vector[1]
            pq.Q(all) | pq.StateVector([2, 1, 0]) * state_vector[2]

        phase_shifter_phis = weights[:3]
        thetas = weights[3:6]
        phis = weights[6:]
        with pq.Program() as interferometer:
            for i in range(3):
                pq.Q(i) | pq.Phaseshifter(phase_shifter_phis[i])

            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[0], phi=phis[0])
            pq.Q(0, 1) | pq.Beamsplitter(theta=thetas[1], phi=phis[1])
            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[2], phi=phis[2])

        with pq.Program() as program:
            pq.Q(all) | preparation

            pq.Q(all) | interferometer

            pq.Q(all) | pq.ImperfectPostSelectPhotons(
                postselect_modes=(1, 2),
                photon_counts=(1, 0),
                detector_efficiency_matrix=np.array(
                    [
                        [1.0, 0.2, 0.1],
                        [0.0, 0.8, 0.2],
                        [0.0, 0.0, 0.7],
                    ]
                ),
            )

        simulator = SimulatorClass(d=3, config=pq.Config(cutoff=4), connector=connector)

        state = simulator.execute(program).state

        norm = state.norm

        density_matrix = state.density_matrix[:3, :3] / norm

        expected_state = np.copy(state_vector)
        expected_state = connector.assign(expected_state, 2, -expected_state[2])

        loss = 1 - np.sqrt(
            np.real(np.conj(expected_state) @ density_matrix @ expected_state)
        )

        return loss

    weights = np.array(
        [np.pi, 0.0, 0.0, np.pi / 8, 65.5302 * 2 * np.pi / 360, -np.pi / 8, 0, 0, 0]
    )

    loss = _calculate_loss(weights=weights)

    loss_gradient = grad(_calculate_loss)

    gradient = loss_gradient(weights)

    assert np.isclose(loss, 0.0485518312033707)

    assert np.allclose(
        gradient,
        [
            0.0,
            0.0,
            0.0,
            0.08767547,
            0.01909362,
            -0.01343052,
            0.0,
            0.0,
            0.0,
        ],
        atol=1e-7,
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import piquasso as pq

from jax import jit, jacfwd


def test_jit_compilation():
    def func(r, theta, xi):
        simulator = pq.PureFockSimulator(
            d=2,
            config=pq.Config(cutoff=5, dtype=np.float32),
            connector=pq.JaxConnector(),
        )

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r)

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta)

            pq.Q(0) | pq.Kerr(xi=xi)

        return simulator.execute(program).state.fock_probabilities

    r = 0.1
    theta = np.pi / 5
    xi = 0.01

    compiled_func = jit(func)

    result = func(r, theta, xi)
    compiled_result = compiled_func(r, theta, xi)

    assert np.allclose(result, compiled_result)


def test_jit_and_jacfwd_compilation():
    def func(r, theta, xi):
        simulator = pq.PureFockSimulator(
            d=2,
            config=pq.Config(cutoff=5, dtype=np.float32),
            connector=pq.JaxConnector(),
        )

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r)

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta)

            pq.Q(0) | pq.Kerr(xi=xi)

        return simulator.execute(program).state.fock_probabilities

    r = 0.1
    theta = np.pi / 5
    xi = 0.01

    jacobian = jacfwd(func, argnums=(0, 1, 2))

    compiled_jacobian = jit(jacobian)

    result = jacobian(r, theta, xi)
    compiled_result = compiled_jacobian(r, theta, xi)

    assert np.allclose(result[0], compiled_result[0])
    assert np.allclose(result[1], compiled_result[1])
    assert np.allclose(result[2], compiled_result[2])



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq
from piquasso._math.fock import cutoff_fock_space_dim
import tensorflow as tf
import time
import numpy as np


PROFILE = False


def create_layer_parameters(d: int):
    number_of_beamsplitters: int
    if d % 2 == 0:
        number_of_beamsplitters = (d // 2) ** 2
        number_of_beamsplitters += ((d - 1) // 2) * (d // 2)
    else:
        number_of_beamsplitters = ((d - 1) // 2) * d

    thetas_1 = [tf.Variable(0.1) for _ in range(number_of_beamsplitters)]
    phis_1 = [tf.Variable(0.0) for _ in range(d - 1)]

    squeezings = [tf.Variable(0.1) for _ in range(d)]

    thetas_2 = [tf.Variable(0.1) for _ in range(number_of_beamsplitters)]
    phis_2 = [tf.Variable(0.0) for _ in range(d - 1)]

    displacements = [tf.Variable(0.1) for _ in range(d)]

    kappas = [tf.Variable(0.1) for _ in range(d)]

    return {
        "d": d,
        "thetas_1": thetas_1,
        "phis_1": phis_1,
        "squeezings": squeezings,
        "thetas_2": thetas_2,
        "phis_2": phis_2,
        "displacements": displacements,
        "kappas": kappas,
    }


input = 0.01
d = 2
cutoff = 10

target_state_vector = np.zeros(cutoff_fock_space_dim(cutoff=cutoff, d=d), dtype=complex)
target_state_vector[1] = 1.0
target_state = tf.constant(target_state_vector, dtype=tf.complex128)

simulator = pq.PureFockSimulator(
    d=d, config=pq.Config(cutoff=cutoff), connector=pq.TensorflowConnector()
)

parameters = create_layer_parameters(d)

if PROFILE:
    profiler_options = tf.profiler.experimental.ProfilerOptions(python_tracer_level=1)
    tf.profiler.experimental.start("logdir", options=profiler_options)


with tf.GradientTape() as tape:
    with pq.Program() as program:
        pq.Q(all) | pq.Vacuum()

        for mode in range(d):
            pq.Q(mode) | pq.Displacement(r=input)

        i = 0
        for col in range(d):
            if col % 2 == 0:
                for mode in range(0, d - 1, 2):
                    modes = (mode, mode + 1)
                    pq.Q(*modes) | pq.Beamsplitter(parameters["thetas_1"][i], phi=0.0)
                    i += 1

            if col % 2 == 1:
                for mode in range(1, d - 1, 2):
                    modes = (mode, mode + 1)
                    pq.Q(*modes) | pq.Beamsplitter(parameters["thetas_1"][i], phi=0.0)
                    i += 1

        for i in range(d - 1):
            pq.Q(i) | pq.Phaseshifter(parameters["phis_1"][i])

        for mode, r in enumerate(parameters["squeezings"]):
            pq.Q(mode) | pq.Squeezing(r)

        i = 0
        for col in range(d):
            if col % 2 == 0:
                for mode in range(0, d - 1, 2):
                    modes = (mode, mode + 1)
                    pq.Q(*modes) | pq.Beamsplitter(parameters["thetas_2"][i], phi=0.0)
                    i += 1

            if col % 2 == 1:
                for mode in range(1, d - 1, 2):
                    modes = (mode, mode + 1)
                    pq.Q(*modes) | pq.Beamsplitter(parameters["thetas_2"][i], phi=0.0)
                    i += 1

        for i in range(d - 1):
            pq.Q(i) | pq.Phaseshifter(parameters["phis_2"][i])

        for mode, r in enumerate(parameters["displacements"]):
            pq.Q(mode) | pq.Displacement(r=r)

        for mode, xi in enumerate(parameters["kappas"]):
            pq.Q(mode) | pq.Kerr(xi=xi)

    start_time = time.time()
    state = simulator.execute(program).state
    print("EXECUTION TIME: ", time.time() - start_time)

    state_vector = state.state_vector

    cost = tf.reduce_sum(tf.abs(target_state_vector - state_vector))


flattened_parameters = (
    parameters["thetas_1"]
    + parameters["phis_1"]
    + parameters["squeezings"]
    + parameters["thetas_2"]
    + parameters["phis_2"]
    + parameters["displacements"]
    + parameters["kappas"]
)

start_time = time.time()
gradient = tape.gradient(cost, flattened_parameters)

print("GRADIENT CALCULATION TIME:", time.time() - start_time)
print("GRADIENT:", [g.numpy() for g in gradient])

if PROFILE:
    tf.profiler.experimental.stop()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Benchmarking the Piquasso vs. TheWalrus hafnian with repetitions.
"""

import time
import json
import numpy as np
from scipy.linalg import schur
import matplotlib.pyplot as plt

from piquasso._math.pfaffian import pfaffian


def pfaffian_naive(matrix):
    if matrix.shape[0] == 0:
        return 1.0

    if matrix.shape[0] % 2 == 1:
        return 0.0

    blocks, O = schur(matrix)
    a = np.diag(blocks, 1)[::2]

    return np.prod(a) * np.linalg.det(O)


np.set_printoptions(suppress=True, linewidth=200)

if __name__ == "__main__":
    x = []
    y = []
    z = []

    ITER = 100

    FILENAME = f"pfaffian_benchmark_{int(time.time())}.json"

    # Compilation
    d = 2
    A = np.random.rand(d, d)
    A = A - A.T

    pfaffian(A)

    for d in range(2, 100 + 1, 2):
        print(d)
        x.append(d)
        A = np.random.rand(d, d)

        A = A - A.T

        sum_ = 0.0
        for _ in range(ITER):
            print("|", end="", flush=True)
            start_time = time.time()
            pfaffian(A)
            sum_ += time.time() - start_time

        y.append(sum_ / ITER)
        print()

        sum_ = 0.0
        for _ in range(ITER):
            print("/", end="", flush=True)
            start_time = time.time()
            pfaffian_naive(A)
            sum_ += time.time() - start_time

        z.append(sum_ / ITER)
        print()

        with open(FILENAME, "w") as f:
            json.dump(dict(x=x, y=y, z=z), f, indent=4)

    plt.scatter(x, y, label="Piquasso C++")
    plt.scatter(x, z, label="Piquasso Python")
    plt.legend()
    plt.xlabel("d [-]")
    plt.ylabel("Execution time [s]")

    plt.show()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Tuple

import numpy as np
import numba as nb

from piquasso._math.combinatorics import comb

from piquasso.api.config import Config
from piquasso._math.combinatorics import partitions
from piquasso._math.gradients import (
    create_single_mode_displacement_gradient,
    create_single_mode_squeezing_gradient,
)
from piquasso._math.gate_matrices import (
    create_single_mode_displacement_matrix,
    create_single_mode_squeezing_matrix,
)
from piquasso._math.indices import get_index_in_fock_space

from piquasso.api.connector import BaseConnector


@nb.njit(cache=True)
def cutoff_fock_space_dim(cutoff: int, d: int) -> int:
    r"""
    Calculates the dimension of the cutoff Fock space with the relation

    ..math::
        \sum_{i=0}^{c - 1} {d + i - 1 \choose i} = {d + c - 1 \choose d}.
    """
    return comb(d + cutoff - 1, d)


@nb.njit(cache=True)
def cutoff_fock_space_dim_array(cutoff, d):
    ret = np.empty(cutoff.shape, dtype=np.int32)

    for i in range(len(cutoff)):
        ret[i] = comb(d + cutoff[i] - 1, d)

    return ret


@nb.njit(cache=True)
def symmetric_subspace_cardinality(d: int, n: int) -> int:
    return comb(d + n - 1, n)


@nb.njit(cache=True)
def nb_get_fock_space_basis(d: int, cutoff: int) -> np.ndarray:
    size = cutoff_fock_space_dim(cutoff=cutoff, d=d)

    ret = np.empty((size, d), dtype=np.int32)
    current_row = 0
    for n in range(cutoff):
        num_rows = symmetric_subspace_cardinality(d, n)
        out = ret[current_row : current_row + num_rows, :]
        _ = partitions(boxes=d, particles=n, out=out)
        current_row += num_rows

    return ret


get_fock_space_basis = functools.lru_cache(maxsize=None)(nb_get_fock_space_basis)


def get_single_mode_squeezing_operator(
    r: float,
    phi: float,
    cutoff: int,
    complex_dtype: np.dtype,
    connector: BaseConnector,
) -> np.ndarray:
    @connector.custom_gradient
    def _single_mode_squeezing_operator(r, phi):
        r = connector.preprocess_input_for_custom_gradient(r)
        phi = connector.preprocess_input_for_custom_gradient(phi)

        matrix = create_single_mode_squeezing_matrix(
            r,
            phi,
            cutoff,
            complex_dtype=complex_dtype,
            connector=connector,
        )
        grad = create_single_mode_squeezing_gradient(
            r,
            phi,
            cutoff,
            matrix,
            connector,
        )
        return matrix, grad

    return _single_mode_squeezing_operator(r, phi)


def get_single_mode_cubic_phase_operator(
    gamma: float, cutoff: int, hbar: float, connector: BaseConnector
) -> np.ndarray:
    r"""Cubic Phase gate.

    The definition of the Cubic Phase gate is

    .. math::
        \operatorname{CP}(\gamma) = e^{i \hat{x}^3 \frac{\gamma}{3 \hbar}}

    The Cubic Phase gate transforms the annihilation operator as

    .. math::
        \operatorname{CP}^\dagger(\gamma) \hat{a} \operatorname{CP}(\gamma) =
            \hat{a} + i\frac{\gamma(\hat{a} +\hat{a}^\dagger)^2}{2\sqrt{2/\hbar}}

    It transforms the :math:`\hat{p}` quadrature as follows:

    .. math::
        \operatorname{CP}^\dagger(\gamma) \hat{p} \operatorname{CP}(\gamma) =
            \hat{p} + \gamma \hat{x}^2.

    Args:
        gamma (float): The Cubic Phase parameter.
        hbar (float): Scaling parameter.
    Returns:
        np.ndarray:
            The resulting transformation, which could be applied to the state.
    """

    np = connector.np

    annih = np.diag(np.sqrt(np.arange(1, cutoff)), 1)
    position = (annih.T + annih) * np.sqrt(hbar / 2)
    return connector.expm(1j * connector.powm(position, 3) * (gamma / (3 * hbar)))


def operator_basis(space):
    for index, basis in enumerate(space):
        for dual_index, dual_basis in enumerate(space):
            yield (index, dual_index), (basis, dual_basis)


def get_creation_operator(
    modes: Tuple[int, ...], space: np.ndarray, config: Config
) -> np.ndarray:
    d = len(space[0])
    size = cutoff_fock_space_dim(cutoff=config.cutoff, d=d)
    operator = np.zeros(shape=(size,) * 2, dtype=config.complex_dtype)

    for index, basis in enumerate(space):
        basis = np.array(basis)
        basis[modes,] += np.ones_like(modes)
        dual_index = get_index_in_fock_space(tuple(basis))

        if dual_index < size:
            operator[dual_index, index] = 1

    return operator


def get_annihilation_operator(
    modes: Tuple[int, ...], space: np.ndarray, config: Config
) -> np.ndarray:
    return get_creation_operator(modes, space, config).transpose()


def get_single_mode_displacement_operator(r, phi, cutoff, complex_dtype, connector):
    @connector.custom_gradient
    def _single_mode_displacement_operator(r, phi):
        r = connector.preprocess_input_for_custom_gradient(r)
        phi = connector.preprocess_input_for_custom_gradient(phi)

        matrix = create_single_mode_displacement_matrix(
            r,
            phi,
            cutoff,
            complex_dtype,
            connector=connector,
        )
        grad = create_single_mode_displacement_gradient(
            r,
            phi,
            cutoff,
            matrix,
            connector,
        )
        return matrix, grad

    return _single_mode_displacement_operator(r, phi)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_Config_seed_generates_same_output():
    seed_sequence = 42

    mean = np.array([1, 2])
    covariance = np.array(
        [
            [2, -1],
            [-1, 2],
        ]
    )

    config1 = pq.Config(seed_sequence=seed_sequence)
    config2 = pq.Config(seed_sequence=seed_sequence)

    sample = config1.rng.multivariate_normal(mean=mean, cov=covariance)
    reproduced_sample = config2.rng.multivariate_normal(mean=mean, cov=covariance)

    assert np.allclose(sample, reproduced_sample)


def test_eq():
    config1 = pq.Config(seed_sequence=1, cutoff=7)
    config2 = pq.Config(seed_sequence=1, cutoff=7)
    config3 = pq.Config(
        seed_sequence=0,
        cache_size=64,
        hbar=1,
        use_torontonian=True,
        cutoff=6,
        measurement_cutoff=4,
        validate=False,
        use_dask=True,
    )

    assert config1 == config1
    assert config3 == config3
    assert config1 == config2
    assert not (config1 == config3)


def test_as_code_dtype():
    conf_f32 = pq.Config(dtype=np.float32)
    conf_f64 = pq.Config(dtype=np.float64)
    conf_f = pq.Config(dtype=float)

    as_code_f32 = conf_f32._as_code()
    as_code_f64 = conf_f64._as_code()
    as_code_f = conf_f._as_code()

    assert as_code_f32 == "pq.Config(dtype=np.float32)"
    assert as_code_f64 == "pq.Config()"
    assert as_code_f == "pq.Config()"


def test_as_code_validate():
    default = pq.Config()
    default_2 = pq.Config(validate=True)
    no_validate_config = pq.Config(validate=False)

    assert default._as_code() == default_2._as_code()
    assert default._as_code() == "pq.Config()"
    assert no_validate_config._as_code() == "pq.Config(validate=False)"


def test_as_code_use_dask():
    default = pq.Config()
    conf_dont_use_dask = pq.Config(use_dask=False)
    conf_use_dask = pq.Config(use_dask=True)

    assert conf_dont_use_dask._as_code() == default._as_code()
    assert conf_use_dask._as_code() == "pq.Config(use_dask=True)"


def test_complex_dtype():
    conf_f32 = pq.Config(dtype=np.float32)
    conf_f64 = pq.Config(dtype=np.float64)
    conf_f = pq.Config(dtype=float)

    assert conf_f32.complex_dtype == np.complex64
    assert conf_f64.complex_dtype == np.complex128
    assert conf_f.complex_dtype == np.complex128


def test_Config_rng_is_shallow_copied():
    config = pq.Config(seed_sequence=123)

    config_copy = config.copy()

    assert config.rng is config_copy.rng



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq

import numpy as np

from jax import jit, grad, jacrev


def test_Beamsplitter_mean_particle_numbers_grad():
    connector = pq.JaxConnector()

    def get_mean_particle_number_first_mode(theta):
        with pq.Program() as program:
            pq.Q() | pq.StateVector([1, 0])

            pq.Q() | pq.Beamsplitter(theta=theta)

        simulator = pq.fermionic.GaussianSimulator(
            d=2, connector=connector, config=pq.Config(validate=False)
        )

        state = simulator.execute(program).state

        return state.mean_particle_numbers(modes=(0,))[0]

    jitted_func = jit(get_mean_particle_number_first_mode)

    theta = np.pi / 3
    result = jitted_func(theta)

    assert np.isclose(result, np.cos(theta) ** 2)

    grad_func = grad(jitted_func)
    result_grad = grad_func(theta)

    assert np.isclose(result_grad, -2 * np.cos(theta) * np.sin(theta))


def test_Beamsplitter_mean_particle_numbers_jacrev():
    connector = pq.JaxConnector()

    def get_mean_particle_numbers(theta):
        with pq.Program() as program:
            pq.Q() | pq.StateVector([1, 0])

            pq.Q() | pq.Beamsplitter(theta=theta)

        simulator = pq.fermionic.GaussianSimulator(
            d=2, connector=connector, config=pq.Config(validate=False)
        )

        state = simulator.execute(program).state

        return state.mean_particle_numbers(modes=(0, 1))

    jitted_func = jit(get_mean_particle_numbers)

    theta = np.pi / 3
    result = jitted_func(theta)

    assert np.isclose(result[0], np.cos(theta) ** 2)
    assert np.isclose(result[1], np.sin(theta) ** 2)

    jac_func = jacrev(jitted_func)
    result_jac = jac_func(theta)

    assert np.isclose(result_jac[0], -2 * np.cos(theta) * np.sin(theta))
    assert np.isclose(result_jac[1], 2 * np.cos(theta) * np.sin(theta))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from piquasso._math.indices import (
    get_index_in_fock_space,
    get_index_in_fock_subspace,
)


def get_index_in_fock_space_for_0_particles():
    assert get_index_in_fock_space((0, 0, 0)) == 0


@pytest.mark.parametrize(
    "vector, index",
    [
        ((0, 0, 1), 3),
        ((0, 1, 0), 2),
        ((1, 0, 0), 1),
    ],
)
def get_index_in_fock_space_for_1_particle(vector, index):
    assert get_index_in_fock_space(vector) == index


@pytest.mark.parametrize(
    "vector, index",
    [
        ((0, 0, 2), 9),
        ((0, 1, 1), 8),
        ((0, 2, 0), 7),
        ((1, 0, 1), 6),
        ((1, 1, 0), 5),
        ((2, 0, 0), 4),
    ],
)
def test_get_index_in_fock_space_for_2_particles(vector, index):
    assert get_index_in_fock_space(vector) == index


@pytest.mark.parametrize(
    "vector, index",
    [
        ((0, 0, 3), 19),
        ((0, 1, 2), 18),
        ((0, 2, 1), 17),
        ((0, 3, 0), 16),
        ((1, 0, 2), 15),
        ((1, 1, 1), 14),
        ((1, 2, 0), 13),
        ((2, 0, 1), 12),
        ((2, 1, 0), 11),
        ((3, 0, 0), 10),
    ],
)
def test_get_index_in_fock_space_for_3_particles(vector, index):
    assert get_index_in_fock_space(vector) == index


def test_get_index_in_fock_subspace_for_0_particles():
    assert get_index_in_fock_subspace((0, 0, 0)) == 0


@pytest.mark.parametrize(
    "vector, index",
    [
        ((0, 0, 1), 2),
        ((0, 1, 0), 1),
        ((1, 0, 0), 0),
    ],
)
def test_get_index_in_fock_subspace_for_1_particle(vector, index):
    assert get_index_in_fock_subspace(vector) == index


@pytest.mark.parametrize(
    "vector, index",
    [
        ((0, 0, 2), 5),
        ((0, 1, 1), 4),
        ((0, 2, 0), 3),
        ((1, 0, 1), 2),
        ((1, 1, 0), 1),
        ((2, 0, 0), 0),
    ],
)
def test_get_index_in_fock_subspace_for_2_particles(vector, index):
    assert get_index_in_fock_subspace(vector) == index


@pytest.mark.parametrize(
    "vector, index",
    [
        ((0, 0, 3), 9),
        ((0, 1, 2), 8),
        ((0, 2, 1), 7),
        ((0, 3, 0), 6),
        ((1, 0, 2), 5),
        ((1, 1, 1), 4),
        ((1, 2, 0), 3),
        ((2, 0, 1), 2),
        ((2, 1, 0), 1),
        ((3, 0, 0), 0),
    ],
)
def test_get_index_in_fock_subspace_for_3_particles(vector, index):
    assert get_index_in_fock_subspace(vector) == index



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, List
import numpy as np

from piquasso._math.fock import cutoff_fock_space_dim
from piquasso._math.linalg import is_unitary

from piquasso.api.config import Config
from piquasso.api.exceptions import InvalidState
from piquasso.api.state import State
from piquasso.api.exceptions import PiquassoException
from piquasso.api.connector import BaseConnector

from .utils import (
    calculate_state_vector,
    calculate_inner_product,
)


class SamplingState(State):
    """A state dedicated for Boson Sampling (or related) simulations.

    When using :class:`~piquasso._simulators.sampling.simulator.SamplingSimulator`, the
    simulation results will contain an instance of this class, containing the input
    occupation numbers and the interferometer to be applied.

    Example usage:

    .. code-block:: python

        >>> import piquasso as pq
        >>>
        >>> from scipy.stats import unitary_group
        >>>
        >>> d = 7
        >>>
        >>> interferometer_matrix = unitary_group.rvs(d)
        >>>
        >>> with pq.Program() as program:
        >>>     pq.Q(all) | pq.StateVector([1, 1, 1, 0, 0, 0, 0])
        >>>     pq.Q(all) | pq.Interferometer(interferometer_matrix)
        >>>     pq.Q(all) | pq.ParticleNumberMeasurement()
        >>>
        >>> simulator = pq.SamplingSimulator(d=d)
        >>>
        >>> result = simulator.execute(program, shots=3)
        >>>
        >>> result.samples
        [(0, 0, 0, 0, 2, 1, 0), (3, 0, 0, 0, 0, 0, 0), (0, 0, 0, 0, 0, 2, 1)]
    """

    def __init__(
        self, d: int, connector: BaseConnector, config: Optional[Config] = None
    ) -> None:
        """
        Args:
            d (int): The number of modes.
            connector (BaseConnector): Instance containing calculation functions.
            config (Config): Instance containing constants for the simulation.
        """
        super().__init__(connector=connector, config=config)

        self._d = d

        self._occupation_numbers: List = []
        self._coefficients: List = []

        self.interferometer = np.diag(np.ones(d, dtype=self._config.complex_dtype))
        """The interferometer matrix corresponding to the circuit."""

        self.is_lossy = False
        """Returns `True` if the state is lossy, otherwise `False`."""

    def validate(self) -> None:
        """Validates the current state.

        Raises:
            InvalidState: If the interferometer matrix is non-unitary, or the input
                state is invalid.
        """
        if not self._config.validate:
            return

        if not is_unitary(self.interferometer):
            raise InvalidState("The interferometer matrix is not unitary.")

        for occupation_number in self._occupation_numbers:
            if len(occupation_number) != self.d:
                raise InvalidState(
                    f"The occupation number '{occupation_number}' is not well-defined "
                    f"on '{self.d}' modes."
                )

        norm = self.norm

        if not np.isclose(norm, 1.0):
            raise InvalidState(f"The state is not normalized: norm={norm}")

    @property
    def d(self):
        return self._d

    @property
    def norm(self):
        """The norm of the state."""
        return np.sum(np.abs(self._coefficients) ** 2)

    def get_particle_detection_probability(
        self, occupation_number: np.ndarray
    ) -> float:
        if self._config.validate and len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        np = self._connector.np
        fallback_np = self._connector.fallback_np

        output_number_of_particles = fallback_np.sum(occupation_number)

        sum_ = 0.0

        for index, input_occupation_number in enumerate(self._occupation_numbers):
            if output_number_of_particles != fallback_np.sum(input_occupation_number):
                continue

            inner_product = calculate_inner_product(
                interferometer=self.interferometer,
                input=input_occupation_number,
                output=fallback_np.array(occupation_number),
                connector=self._connector,
            )
            coefficient = self._coefficients[index]

            sum_ += coefficient * inner_product

        return np.abs(sum_) ** 2

    @property
    def state_vector(self):
        """The state vector representation of this state.

        This implementation follows Algorithm 1 (`SLOS_full`) from
        `Strong Simulation of Linear Optical Processes <https://arxiv.org/pdf/2206.10549>`_.
        """  # noqa: E501

        connector = self._connector
        np = connector.np
        fallback_np = connector.fallback_np

        particle_numbers = fallback_np.sum(self._occupation_numbers, axis=1)

        state_vector = np.zeros(
            shape=cutoff_fock_space_dim(d=self.d, cutoff=self._config.cutoff),
            dtype=self._config.complex_dtype,
        )

        for index in range(len(self._occupation_numbers)):
            particle_number = particle_numbers[index]

            starting_index = cutoff_fock_space_dim(d=self.d, cutoff=particle_number)

            ending_index = cutoff_fock_space_dim(d=self.d, cutoff=particle_number + 1)

            coefficient = self._coefficients[index]

            input_state = self._occupation_numbers[index]

            partial_state_vector = calculate_state_vector(
                self.interferometer, input_state, self._config, self._connector
            )

            index_range = fallback_np.arange(starting_index, ending_index)

            state_vector = connector.assign(
                state_vector,
                index_range,
                state_vector[index_range] + coefficient * partial_state_vector,
            )

        return state_vector

    @property
    def density_matrix(self) -> np.ndarray:
        """The density matrix of the state in the truncated Fock space."""
        state_vector = self.state_vector

        return self._np.outer(state_vector, self._np.conj(state_vector))

    @property
    def fock_probabilities(self) -> np.ndarray:
        np = self._connector.np

        state_vector = self.state_vector

        return np.real(np.conj(state_vector) * state_vector)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SamplingState):
            return False

        return (
            np.allclose(self._coefficients, other._coefficients)
            and np.allclose(self._occupation_numbers, other._occupation_numbers)
            and np.allclose(self.interferometer, other.interferometer)
            and self.is_lossy == other.is_lossy
        )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


@pytest.fixture
def state():
    with pq.Program() as initialization:
        pq.Q(0) | pq.Displacement(r=1)
        pq.Q(1) | pq.Displacement(r=1, phi=np.pi / 2)
        pq.Q(2) | pq.Displacement(r=1, phi=np.pi / 4)

        pq.Q(0) | pq.Squeezing(np.log(2), phi=np.pi / 2)
        pq.Q(1) | pq.Squeezing(np.log(1), phi=np.pi / 4)
        pq.Q(2) | pq.Squeezing(np.log(2), phi=np.pi / 2)

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(initialization).state
    state.validate()

    return state



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


@pytest.fixture
def d():
    return 3


@pytest.fixture
def state(d):
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=2, phi=np.pi / 3)
        pq.Q(1) | pq.Displacement(r=1, phi=np.pi / 4)
        pq.Q(2) | pq.Displacement(r=1 / 2, phi=np.pi / 6)

        pq.Q(1) | pq.Squeezing(r=1 / 2, phi=np.pi / 4)
        pq.Q(2) | pq.Squeezing(r=3 / 4)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state

    return state


@pytest.fixture
def nondisplaced_state(d):
    with pq.Program() as program:
        pq.Q(1) | pq.Squeezing(r=1 / 2, phi=np.pi / 4)
        pq.Q(2) | pq.Squeezing(r=3 / 4)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state

    return state


def test_measure_homodyne_zeroes_state_on_measured_modes(state):
    with pq.Program() as program:
        pq.Q(0) | pq.HomodyneMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    assert state.xpxp_mean_vector[0] == 0
    assert state.xpxp_covariance_matrix[0][0] == state._config.hbar


def test_measure_homodyne_with_angle_does_not_alter_the_state(state):
    angle = np.pi / 3

    with pq.Program() as program:
        pq.Q(0) | pq.HomodyneMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    measured_state = simulator.execute(program, initial_state=state).state
    measured_state.validate()

    with pq.Program() as program_with_rotation:
        pq.Q(0) | pq.HomodyneMeasurement(angle)

    state_with_rotation = simulator.execute(
        program_with_rotation, initial_state=state
    ).state
    state_with_rotation.validate()

    assert measured_state == state_with_rotation


def test_measure_homodyne_with_multiple_shots(state):
    shots = 4

    with pq.Program() as program:
        pq.Q(0, 1) | pq.HomodyneMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    result = simulator.execute(program, initial_state=state, shots=shots)

    assert len(result.samples) == shots


def test_measure_heterodyne_zeroes_state_on_measured_modes(state):
    with pq.Program() as program:
        pq.Q(0) | pq.HeterodyneMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    assert state.xpxp_mean_vector[0] == 0
    assert state.xpxp_covariance_matrix[0][0] == state._config.hbar


def test_measure_heterodyne_with_multiple_shots(state):
    shots = 4

    with pq.Program() as program:
        pq.Q(0, 1) | pq.HeterodyneMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    result = simulator.execute(program, initial_state=state, shots=shots)

    assert len(result.samples) == shots


def test_measure_dyne(state):
    detection_covariance = np.array(
        [
            [2, 0],
            [0, 0.5],
        ]
    )

    with pq.Program() as program:
        pq.Q(0) | pq.GeneraldyneMeasurement(detection_covariance)

    simulator = pq.GaussianSimulator(d=state.d)
    evolved_state = simulator.execute(program, initial_state=state).state

    evolved_state.validate()


def test_measure_dyne_with_multiple_shots(state):
    shots = 4

    detection_covariance = np.array(
        [
            [2, 0],
            [0, 0.5],
        ]
    )

    with pq.Program() as program:
        pq.Q(0) | pq.GeneraldyneMeasurement(detection_covariance)

    simulator = pq.GaussianSimulator(d=state.d)
    result = simulator.execute(program, initial_state=state, shots=shots)

    assert len(result.samples) == shots


def test_measure_particle_number_on_one_modes(state):
    with pq.Program() as program:
        pq.Q(0) | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    result = simulator.execute(program, initial_state=state)

    assert result


def test_measure_particle_number_on_two_modes(state):
    with pq.Program() as program:
        pq.Q(0, 1) | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    result = simulator.execute(program, initial_state=state)

    assert result


def test_measure_particle_number_on_all_modes(state):
    with pq.Program() as program:
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)
    result = simulator.execute(program, initial_state=state)

    assert result


@pytest.mark.parametrize(
    "MeasurementClass",
    (pq.HomodyneMeasurement, pq.HeterodyneMeasurement),
)
def test_dyneMeasurement_resulting_state_inherits_config(MeasurementClass):
    with pq.Program() as program:
        pq.Q(0) | pq.Fourier()
        pq.Q(0, 2) | pq.Squeezing2(r=0.5, phi=0)
        pq.Q(0, 1) | pq.Beamsplitter(theta=1, phi=0)
        pq.Q(0) | MeasurementClass()

    custom_config = pq.Config(hbar=3, seed_sequence=42)

    simulator = pq.GaussianSimulator(d=3, config=custom_config)

    state = simulator.execute(program).state

    assert state._config == custom_config


@pytest.mark.parametrize(
    "MeasurementClass",
    (pq.HomodyneMeasurement, pq.HeterodyneMeasurement),
)
def test_dyneMeasurement_results_in_same_state_regardless_of_hbar(MeasurementClass):
    seed_sequence = 42

    with pq.Program() as program:
        pq.Q(0) | pq.Fourier()
        pq.Q(0, 2) | pq.Squeezing2(r=0.5, phi=0)
        pq.Q(0, 1) | pq.Beamsplitter(theta=1, phi=0)
        pq.Q(0) | MeasurementClass()

    simulator_hbar_1 = pq.GaussianSimulator(
        d=3, config=pq.Config(hbar=1, seed_sequence=seed_sequence)
    )
    simulator_hbar_3 = pq.GaussianSimulator(
        d=3, config=pq.Config(hbar=3, seed_sequence=seed_sequence)
    )

    state_hbar_1 = simulator_hbar_1.execute(program).state
    state_hbar_3 = simulator_hbar_3.execute(program).state

    assert state_hbar_1 == state_hbar_3


def test_displaced_ThresholdMeasurement_with_torontonian(state):
    state._config.use_torontonian = True

    state_with_nonzero_displacements = state

    with pq.Program() as program:
        pq.Q(0) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(d=state.d)

    simulator.execute(program, initial_state=state_with_nonzero_displacements)


def test_measure_threshold_on_one_modes(nondisplaced_state):
    with pq.Program() as program:
        pq.Q(0) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(d=nondisplaced_state.d)
    result = simulator.execute(program, initial_state=nondisplaced_state)

    assert result


def test_measure_threshold_with_multiple_shots(nondisplaced_state):
    with pq.Program() as program:
        pq.Q(0) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(d=nondisplaced_state.d)
    result = simulator.execute(program, initial_state=nondisplaced_state, shots=20)

    assert result


def test_measure_threshold_on_two_modes(nondisplaced_state):
    with pq.Program() as program:
        pq.Q(0, 1) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(d=nondisplaced_state.d)
    result = simulator.execute(program, initial_state=nondisplaced_state)

    assert result


def test_measure_threshold_on_all_modes(nondisplaced_state):
    with pq.Program() as program:
        pq.Q() | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(d=nondisplaced_state.d)
    result = simulator.execute(program, initial_state=nondisplaced_state)

    assert result


def test_seeded_gaussian_boson_sampling():
    d = 5
    shots = 10

    A = np.array(
        [
            [0, 1, 0, 1, 1],
            [1, 0, 0, 0, 1],
            [0, 0, 0, 1, 0],
            [1, 0, 1, 0, 1],
            [1, 1, 0, 1, 0],
        ]
    )

    with pq.Program() as gaussian_boson_sampling:
        pq.Q(all) | pq.Graph(A)

        pq.Q(all) | pq.ParticleNumberMeasurement()

    simulator1 = pq.GaussianSimulator(
        d=d, connector=pq.NumpyConnector(), config=pq.Config(seed_sequence=123)
    )
    result1 = simulator1.execute(gaussian_boson_sampling, shots=shots)

    simulator2 = pq.GaussianSimulator(
        d=d, connector=pq.NumpyConnector(), config=pq.Config(seed_sequence=123)
    )
    result2 = simulator2.execute(gaussian_boson_sampling, shots=shots)

    assert np.allclose(result1.samples, result2.samples)


@pytest.mark.parametrize("use_dask", (True, False))
def test_seeded_gaussian_boson_sampling_samples_from_graph(use_dask):
    d = 5
    shots = 50

    A = np.array(
        [
            [0, 1, 0, 1, 1],
            [1, 0, 0, 0, 1],
            [0, 0, 0, 1, 0],
            [1, 0, 1, 0, 1],
            [1, 1, 0, 1, 0],
        ]
    )

    with pq.Program() as gaussian_boson_sampling:
        pq.Q(all) | pq.Graph(A)

        pq.Q(all) | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(
        d=d,
        connector=pq.NumpyConnector(),
        config=pq.Config(seed_sequence=123, use_dask=use_dask),
    )
    samples = simulator.execute(gaussian_boson_sampling, shots=shots).samples

    assert np.allclose(
        samples,
        [
            [0, 0, 0, 0, 0],
            [3, 1, 0, 1, 1],
            [3, 0, 0, 2, 1],
            [4, 4, 2, 3, 3],
            [0, 0, 0, 0, 0],
            [4, 4, 0, 3, 3],
            [2, 1, 1, 2, 2],
            [0, 0, 0, 1, 1],
            [0, 0, 0, 1, 1],
            [4, 3, 0, 4, 3],
            [3, 2, 0, 1, 2],
            [2, 0, 0, 1, 1],
            [0, 0, 0, 0, 0],
            [1, 2, 0, 0, 1],
            [2, 3, 0, 1, 2],
            [0, 1, 0, 0, 1],
            [0, 1, 0, 2, 3],
            [3, 2, 0, 1, 2],
            [1, 0, 0, 1, 0],
            [0, 0, 0, 0, 0],
            [4, 3, 1, 3, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [1, 0, 0, 1, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [2, 1, 0, 1, 0],
            [3, 1, 0, 1, 1],
            [0, 0, 0, 0, 0],
            [1, 1, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 1, 1, 0],
            [0, 1, 0, 0, 1],
            [0, 0, 0, 0, 0],
            [0, 1, 0, 0, 1],
            [0, 0, 0, 0, 0],
            [1, 0, 0, 1, 0],
            [0, 1, 0, 0, 1],
            [2, 2, 0, 3, 3],
            [0, 0, 1, 1, 0],
            [1, 2, 0, 1, 4],
            [2, 2, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [1, 0, 0, 0, 1],
            [1, 0, 0, 1, 0],
            [0, 0, 1, 1, 0],
            [2, 0, 0, 1, 1],
            [2, 1, 0, 1, 0],
            [2, 1, 0, 1, 2],
            [1, 1, 0, 0, 0],
        ],
    )


def test_seeded_gaussian_boson_sampling_samples_displaced():
    d = 5
    shots = 50

    with pq.Program() as gaussian_boson_sampling:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(r=0.1) | pq.Displacement(r=1)

        pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
        pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
        pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
        pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
        pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
        pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
        pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
        pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

        pq.Q(0, 1, 2) | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(
        d=d, connector=pq.NumpyConnector(), config=pq.Config(seed_sequence=123)
    )

    samples = simulator.execute(gaussian_boson_sampling, shots=shots).samples

    assert np.allclose(
        samples,
        [
            [1, 1, 1],
            [3, 1, 1],
            [0, 3, 0],
            [0, 1, 0],
            [2, 2, 1],
            [1, 2, 0],
            [0, 2, 2],
            [0, 1, 3],
            [1, 1, 1],
            [0, 4, 1],
            [0, 3, 0],
            [0, 2, 1],
            [0, 4, 0],
            [0, 2, 1],
            [1, 2, 0],
            [0, 1, 1],
            [0, 4, 0],
            [0, 2, 0],
            [1, 2, 0],
            [0, 3, 1],
            [0, 3, 1],
            [0, 0, 1],
            [0, 3, 0],
            [0, 2, 1],
            [0, 0, 3],
            [2, 1, 0],
            [1, 3, 0],
            [1, 3, 0],
            [0, 2, 0],
            [1, 3, 0],
            [0, 3, 0],
            [1, 2, 0],
            [0, 1, 0],
            [0, 4, 0],
            [0, 4, 1],
            [0, 0, 1],
            [0, 3, 1],
            [0, 2, 0],
            [1, 2, 0],
            [0, 4, 0],
            [0, 2, 1],
            [0, 4, 1],
            [0, 2, 1],
            [0, 3, 2],
            [0, 1, 1],
            [0, 1, 0],
            [0, 2, 0],
            [0, 2, 0],
            [0, 1, 0],
            [1, 1, 1],
        ],
    )


def test_ThresholdMeasurement_use_torontonian_seeding():
    d = 5
    shots = 10

    seed_sequence = 123

    A = np.array(
        [
            [0, 1, 0, 1, 1],
            [1, 0, 0, 0, 1],
            [0, 0, 0, 1, 0],
            [1, 0, 1, 0, 1],
            [1, 1, 0, 1, 0],
        ]
    )

    with pq.Program() as program:
        pq.Q(all) | pq.Graph(A)

        pq.Q(all) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(
        d=d,
        connector=pq.NumpyConnector(),
        config=pq.Config(seed_sequence=seed_sequence, use_torontonian=True),
    )
    result = simulator.execute(program, shots=shots)

    assert result.samples == [
        (1, 0, 0, 0, 1),
        (1, 1, 0, 1, 1),
        (1, 0, 1, 1, 1),
        (1, 1, 0, 1, 1),
        (0, 0, 0, 0, 0),
        (0, 0, 0, 1, 1),
        (1, 1, 0, 1, 1),
        (0, 0, 0, 0, 0),
        (0, 0, 0, 0, 0),
        (0, 0, 0, 0, 0),
    ]


def test_ThresholdMeasurement_use_torontonian_seeding_float32():
    d = 5
    shots = 10

    seed_sequence = 123

    A = np.array(
        [
            [0, 1, 0, 1, 1],
            [1, 0, 0, 0, 1],
            [0, 0, 0, 1, 0],
            [1, 0, 1, 0, 1],
            [1, 1, 0, 1, 0],
        ]
    )

    with pq.Program() as program:
        pq.Q(all) | pq.Graph(A)

        pq.Q(all) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(
        d=d,
        connector=pq.NumpyConnector(),
        config=pq.Config(
            seed_sequence=seed_sequence, use_torontonian=True, dtype=np.float32
        ),
    )
    result = simulator.execute(program, shots=shots)

    assert result.samples == [
        (1, 0, 0, 0, 1),
        (1, 1, 0, 1, 1),
        (1, 0, 1, 1, 1),
        (1, 1, 0, 1, 1),
        (0, 0, 0, 0, 0),
        (0, 0, 0, 1, 1),
        (1, 1, 0, 1, 1),
        (0, 0, 0, 0, 0),
        (0, 0, 0, 0, 0),
        (0, 0, 0, 0, 0),
    ]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, List

import scipy
import numpy as np

from itertools import repeat
from functools import lru_cache, partial

from scipy.special import factorial

from .state import GaussianState
from .probabilities import (
    calculate_click_probability_nondisplaced,
    calculate_click_probability,
)

from piquasso.instructions import gates

from piquasso.api.result import Result
from piquasso.api.instruction import Instruction
from piquasso.api.exceptions import InvalidInstruction

from piquasso._math.indices import get_operator_index, get_auxiliary_operator_index
from piquasso._math.decompositions import (
    williamson,
    decompose_adjacency_matrix_into_circuit,
)


def passive_linear(
    state: GaussianState, instruction: gates._PassiveLinearGate, shots: int
) -> Result:
    modes = instruction.modes
    passive_block: np.ndarray = instruction._get_passive_block(
        state._connector, state._config
    )

    _apply_passive_linear(state, passive_block, modes=modes)

    return Result(state=state)


def _apply_passive_linear(state, passive_block, modes):
    connector = state._connector

    state._m = connector.assign(state._m, (modes,), passive_block @ state._m[modes,])

    _apply_passive_linear_to_C_and_G(state, passive_block, modes=modes)


def _apply_passive_linear_to_C_and_G(
    state: GaussianState, T: np.ndarray, modes: Tuple[int, ...]
) -> None:
    connector = state._connector

    index = get_operator_index(modes)

    state._C = connector.assign(
        state._C, index, T.conjugate() @ state._C[index] @ T.transpose()
    )
    state._G = connector.assign(state._G, index, T @ state._G[index] @ T.transpose())

    auxiliary_modes = state._get_auxiliary_modes(modes)

    if len(auxiliary_modes) != 0:
        _apply_passive_linear_to_auxiliary_modes(state, T, modes, auxiliary_modes)


def _apply_passive_linear_to_auxiliary_modes(
    state: GaussianState,
    T: np.ndarray,
    modes: Tuple[int, ...],
    auxiliary_modes: Tuple[int, ...],
) -> None:
    connector = state._connector
    np = connector.np

    auxiliary_index = get_auxiliary_operator_index(modes, auxiliary_modes)

    state._C = connector.assign(
        state._C, auxiliary_index, T.conjugate() @ state._C[auxiliary_index]
    )
    state._G = connector.assign(
        state._G, auxiliary_index, T @ state._G[auxiliary_index]
    )

    assign_index = np.ix_(np.arange(state.d), np.array(modes))
    state._C = connector.assign(
        state._C, assign_index, np.conj(state._C[modes, :]).transpose()
    )
    state._G = connector.assign(state._G, assign_index, state._G[modes, :].transpose())


def linear(
    state: GaussianState, instruction: gates._ActiveLinearGate, shots: int
) -> Result:
    modes = instruction.modes
    passive_block: np.ndarray = instruction._get_passive_block(
        state._connector, state._config
    )
    active_block: np.ndarray = instruction._get_active_block(
        state._connector, state._config
    )

    _apply_linear(state, passive_block, active_block, modes)

    return Result(state=state)


def _apply_linear(state, passive_block, active_block, modes):
    connector = state._connector
    np = connector.np

    passive_part = passive_block @ state._m[(modes,)]

    active_part = active_block @ np.conj(state._m[modes,])

    state._m = connector.assign(state._m, (modes,), passive_part + active_part)

    _apply_linear_to_C_and_G(state, passive_block, active_block, modes)


def _apply_linear_to_C_and_G(
    state: GaussianState, P: np.ndarray, A: np.ndarray, modes: Tuple[int, ...]
) -> None:
    connector = state._connector
    np = connector.np

    index = get_operator_index(modes)

    original_C = state._C[index]
    original_G = state._G[index]

    state._G = connector.assign(
        state._G,
        index,
        P @ original_G @ P.transpose()
        + A @ original_G.conjugate().transpose() @ A.transpose()
        + P @ (original_C.transpose() + np.identity(len(modes))) @ A.transpose()
        + A @ original_C @ P.transpose(),
    )

    state._C = connector.assign(
        state._C,
        index,
        P.conjugate() @ original_C @ P.transpose()
        + A.conjugate()
        @ (original_C.transpose() + np.identity(len(modes)))
        @ A.transpose()
        + P.conjugate() @ original_G.conjugate().transpose() @ A.transpose()
        + A.conjugate() @ original_G @ P.transpose(),
    )

    auxiliary_modes = state._get_auxiliary_modes(modes)

    if len(auxiliary_modes) != 0:
        _apply_linear_to_auxiliary_modes(state, P, A, modes, auxiliary_modes)


def _apply_linear_to_auxiliary_modes(
    state: GaussianState,
    P: np.ndarray,
    A: np.ndarray,
    modes: Tuple[int, ...],
    auxiliary_modes: Tuple[int, ...],
) -> None:
    connector = state._connector
    np = connector.np

    auxiliary_index = get_auxiliary_operator_index(modes, auxiliary_modes)

    auxiliary_C = state._C[auxiliary_index]
    auxiliary_G = state._G[auxiliary_index]

    state._C = connector.assign(
        state._C,
        auxiliary_index,
        P.conjugate() @ auxiliary_C + A.conjugate() @ auxiliary_G,
    )

    state._G = connector.assign(
        state._G, auxiliary_index, P @ auxiliary_G + A @ auxiliary_C
    )

    assign_index = np.ix_(np.arange(state.d), np.array(modes))

    state._C = connector.assign(
        state._C, assign_index, state._C[modes, :].conjugate().transpose()
    )
    state._G = connector.assign(state._G, assign_index, state._G[modes, :].transpose())


def displacement(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    connector = state._connector
    np = connector.np

    modes = instruction.modes
    r = instruction._all_params["r"]
    phi = instruction._all_params["phi"]

    indices = np.ix_(np.array(modes))

    state._m = connector.assign(
        state._m, indices, state._m[indices] + r * np.exp(1j * phi)
    )

    return Result(state=state)


def generaldyne_measurement(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    modes = instruction.modes
    detection_covariance = instruction._all_params["detection_covariance"]

    samples = _get_generaldyne_samples(
        state,
        modes,
        shots,
        detection_covariance,
    )

    # NOTE: We choose the last sample for multiple shots.
    sample = samples[-1]

    d = len(state)

    auxiliary_modes = state._get_auxiliary_modes(modes)
    outer_indices = _map_modes_to_xpxp_indices(auxiliary_modes)

    evolved_state = _get_generaldyne_evolved_state(
        state,
        sample,
        modes,
        detection_covariance,
    )

    evolved_mean = np.zeros(2 * d)
    evolved_mean[outer_indices] = evolved_state.xpxp_mean_vector

    evolved_cov = np.identity(2 * d) * state._config.hbar
    evolved_cov[np.ix_(outer_indices, outer_indices)] = (
        evolved_state.xpxp_covariance_matrix
    )

    state.xpxp_mean_vector = evolved_mean
    state.xpxp_covariance_matrix = evolved_cov

    return Result(state=state, samples=list(map(tuple, list(samples))))


def _get_generaldyne_samples(state, modes, shots, detection_covariance):
    indices = _map_modes_to_xpxp_indices(modes)

    full_detection_covariance = state._config.hbar * scipy.linalg.block_diag(
        *[detection_covariance] * len(modes)
    )

    mean = state.xpxp_mean_vector[indices]

    cov = (
        state.xpxp_covariance_matrix[np.ix_(indices, indices)]
        + full_detection_covariance
    )

    # HACK: We need tol=1e-7 to avoid Numpy warnings at homodyne detection with
    # squeezed detection covariance. Numpy warns
    # 'covariance is not positive-semidefinite.', but it definitely is. In the SVG
    # decomposition (which numpy uses for the multivariate normal distribution)
    # the U^T and V matrices should equal, but our covariance might contain too
    # large values leading to inequality, resulting in warning.
    #
    # We might be better of setting `check_valid='ignore'` and verifying
    # positive-definiteness for ourselves.
    return state._config.rng.multivariate_normal(
        mean=mean,
        cov=cov,
        size=shots,
        tol=1e-7,
    )


def _get_generaldyne_evolved_state(state, sample, modes, detection_covariance):
    full_detection_covariance = state._config.hbar * scipy.linalg.block_diag(
        *[detection_covariance] * len(modes)
    )

    mean = state.xpxp_mean_vector
    cov = state.xpxp_covariance_matrix

    indices = _map_modes_to_xpxp_indices(modes)
    outer_indices = np.delete(np.arange(2 * state.d), indices)

    mean_measured = mean[indices]
    mean_outer = mean[outer_indices]

    cov_measured = cov[np.ix_(indices, indices)]
    cov_outer = cov[np.ix_(outer_indices, outer_indices)]
    cov_correlation = cov[np.ix_(outer_indices, indices)]

    evolved_cov_outer = (
        cov_outer
        - cov_correlation
        @ np.linalg.inv(cov_measured + full_detection_covariance)
        @ cov_correlation.transpose()
    )

    evolved_r_A = mean_outer + cov_correlation @ np.linalg.inv(
        cov_measured + full_detection_covariance
    ) @ (sample - mean_measured)

    new_state = GaussianState(
        d=len(evolved_r_A) // 2, connector=state._connector, config=state._config
    )

    new_state.xpxp_covariance_matrix = evolved_cov_outer
    new_state.xpxp_mean_vector = evolved_r_A

    return new_state


def _map_modes_to_xpxp_indices(modes):
    indices = []

    for mode in modes:
        indices.extend([2 * mode, 2 * mode + 1])

    return indices


def particle_number_measurement(
    state: GaussianState,
    instruction: Instruction,
    shots: int,
) -> Result:
    samples = _get_particle_number_measurement_samples(state, instruction, shots)

    return Result(state=state, samples=samples)


def _get_particle_number_measurement_samples(
    state: GaussianState,
    instruction: Instruction,
    shots: int,
) -> np.ndarray:
    modes: Tuple[int, ...] = instruction.modes
    connector = state._connector
    config = state._config

    reduced_state = state.reduced(modes)

    d = reduced_state.d

    hbar = config.hbar
    hbar_in_calculations = 2.0

    normalized_cov = (
        hbar / (2.0 * hbar_in_calculations) * reduced_state.xxpp_covariance_matrix
    )
    normalized_mean = (
        np.sqrt(hbar / hbar_in_calculations) * reduced_state.xxpp_mean_vector
    )
    S, D = williamson(normalized_cov, connector)

    T = S @ S.T
    mixed_diag = D - 2 * np.identity(len(D)) / 2
    mixed_diag[mixed_diag < 0.0] = 0.0

    I = np.identity(d)
    F = np.block([[I, 1j * I], [I, -1j * I]]) / np.sqrt(2)
    Q = (F @ T @ F.conj().T + np.identity(2 * d)) / 2
    B = -np.linalg.inv(Q)[d:, :d]

    sqrt_cov_1 = S @ np.sqrt(mixed_diag)
    sqrt_cov_2 = np.linalg.cholesky(T + np.identity(2 * d))

    _generate_sample_from_seed = partial(
        _generate_sample,
        B=B,
        sqrt_cov_1=sqrt_cov_1,
        sqrt_cov_2=sqrt_cov_2,
        mean=normalized_mean,
        loop_hafnian_batch=reduced_state._connector.loop_hafnian_batch,
        cutoff=config.measurement_cutoff,
    )

    seed = config.seed_sequence

    if config.use_dask:
        try:
            import dask
        except ImportError:
            raise ImportError("This feature requires 'dask' to be installed.")

        delayed_func = dask.delayed(_generate_sample_from_seed)

        compute_list = []
        for idx in range(shots):
            compute_list.append(delayed_func(seed=seed + idx))

        results = dask.compute(*compute_list)

        return np.vstack(results)

    else:
        samples = np.empty(shape=(shots, d), dtype=int)

        for idx in range(shots):
            sample = _generate_sample_from_seed(seed=seed + idx)

            samples[idx] = sample

        return samples


def _generate_sample(B, sqrt_cov_1, sqrt_cov_2, mean, loop_hafnian_batch, cutoff, seed):
    d = len(B)

    rng = np.random.default_rng(seed)

    possible_choices = np.arange(cutoff)

    sample = np.zeros(d, dtype=int)

    pure_mean = mean + sqrt_cov_1 @ rng.normal(size=2 * d)
    pure_mean_complex = (pure_mean[:d] + 1j * pure_mean[d:]) / 2

    evolved_mean = pure_mean + sqrt_cov_2 @ rng.normal(size=2 * d)
    evolved_mean_complex = (evolved_mean[:d] + 1j * evolved_mean[d:]) / 2
    gamma = pure_mean_complex.conj() + B @ (evolved_mean_complex - pure_mean_complex)

    for mode in range(d):
        gamma -= evolved_mean_complex[mode] * B[:, mode]
        mode_p_1 = mode + 1
        matrix = B[:mode_p_1, :mode_p_1]
        diagonal = gamma[:mode_p_1]
        partial_sample = sample[:mode_p_1]
        lhaf_values = loop_hafnian_batch(matrix, diagonal, partial_sample, cutoff)
        weights = np.abs(lhaf_values) ** 2 / factorial(possible_choices)
        weights /= np.sum(weights)

        sample[mode] = rng.choice(possible_choices, p=weights)

    return sample


def threshold_measurement(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    """
    NOTE: The same logic is used here, as for the particle number measurement.
    However, at threshold measurement there is no sense of measurement cutoff,
    therefore it is set to 2 to make the logic sensible in this case as well.

    Also note, that one could speed up this calculation by not calculating the
    probability of clicks (i.e. 1 as sample), and argue that the probability of a
    click is equal to one minus the probability of no click.
    """
    if state._config.use_torontonian:
        samples = _generate_threshold_samples_using_torontonian(
            state, instruction, shots
        )
    else:
        samples = _generate_threshold_samples_using_hafnian(state, instruction, shots)

    return Result(state=state, samples=samples)


def _generate_threshold_samples_using_torontonian(state, instruction, shots):
    is_displaced = state._is_displaced()
    rng = state._config.rng
    hbar = state._config.hbar

    modes = instruction.modes

    @lru_cache(state._config.cache_size)
    def get_probability(
        subspace_modes: Tuple[int, ...], occupation_numbers: Tuple[int, ...]
    ) -> float:
        reduced_state = state.reduced(subspace_modes)

        if not is_displaced:
            return calculate_click_probability_nondisplaced(
                reduced_state.xpxp_covariance_matrix / hbar,
                occupation_numbers,
            )

        return calculate_click_probability(
            reduced_state.xpxp_covariance_matrix / hbar,
            reduced_state.xpxp_mean_vector / np.sqrt(hbar),
            occupation_numbers,
        )

    samples = []

    for _ in repeat(None, shots):
        sample: List[int] = []

        previous_probability = 1.0

        for mode_index in range(1, len(modes) + 1):
            subspace_modes = tuple(modes[:mode_index])

            occupation_numbers = tuple(sample + [0])

            probability = get_probability(
                subspace_modes=subspace_modes, occupation_numbers=occupation_numbers
            )
            conditional_probability = probability / previous_probability

            choice: int
            guess = rng.uniform()

            if guess < conditional_probability:
                choice = 0
                previous_probability *= conditional_probability
            else:
                choice = 1
                previous_probability *= 1 - conditional_probability

            sample.append(choice)

        samples.append(tuple(sample))

    return samples


def _generate_threshold_samples_using_hafnian(state, instruction, shots):
    samples = _get_particle_number_measurement_samples(state, instruction, shots)

    threshold_samples = []

    for sample in samples:
        threshold_samples.append(
            [1 if photon_number else 0 for photon_number in sample]
        )

    return threshold_samples


def homodyne_measurement(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    phi = instruction._all_params["phi"]

    modes = instruction.modes

    phaseshift = np.identity(len(instruction.modes)) * np.exp(-1j * phi)

    _apply_passive_linear(state, phaseshift, modes=modes)

    result = generaldyne_measurement(state, instruction, shots)

    return Result(state=state, samples=result.samples)


def vacuum(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    state.reset()

    return Result(state=state)


def mean(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    state.xpxp_mean_vector = instruction._all_params["mean"] * np.sqrt(
        state._config.hbar
    )

    return Result(state=state)


def covariance(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    state.xpxp_covariance_matrix = instruction._all_params["cov"] * state._config.hbar

    return Result(state=state)


def graph(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    """
    TODO: Find a better solution for multiple operations.
    """
    squeezings, interferometer = decompose_adjacency_matrix_into_circuit(
        adjacency_matrix=instruction._params["adjacency_matrix"],
        mean_photon_number=instruction._params["mean_photon_number"],
        connector=state._connector,
    )

    for mode, r in zip(instruction.modes, squeezings):
        _apply_linear(
            state=state,
            passive_block=np.array([[np.cosh(r)]]),
            active_block=np.array([[np.sinh(r)]]),
            modes=(mode,),
        )

    _apply_passive_linear(state, interferometer, instruction.modes)

    return Result(state=state)


def deterministic_gaussian_channel(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    X = instruction._all_params["X"]
    Y = instruction._all_params["Y"] * state._config.hbar
    modes = instruction.modes

    if state._config.validate and len(X) != 2 * len(modes) or len(Y) != 2 * len(modes):
        raise InvalidInstruction(
            f"The instruction should be specified for '{len(modes)}' modes: "
            f"instruction={instruction}"
        )

    indices = _map_modes_to_xpxp_indices(modes)
    matrix_indices = np.ix_(indices, indices)

    mean_vector = state.xpxp_mean_vector
    covariance_matrix = state.xpxp_covariance_matrix

    embedded_X = np.identity(len(mean_vector), dtype=state._config.complex_dtype)
    embedded_X[matrix_indices] = X

    embedded_Y = np.zeros_like(embedded_X)
    embedded_Y[matrix_indices] = Y

    state.xpxp_mean_vector = embedded_X @ mean_vector
    state.xpxp_covariance_matrix = (
        embedded_X @ covariance_matrix @ embedded_X.T + embedded_Y
    )

    return Result(state=state)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import builtins

import pytest


@pytest.fixture
def raise_ImportError_when_importing_jax():
    real_import = builtins.__import__

    def fake_import(name, globals, locals, fromlist, level):
        if name == "jax":
            raise ImportError()

        return real_import(name, globals, locals, fromlist, level)

    builtins.__import__ = fake_import

    yield

    builtins.__import__ = real_import


@pytest.fixture
def unimport_jax():
    """
    Deletes `jax` from `sys.modules`.
    """

    if "jax" in sys.modules:
        del sys.modules["jax"]


def test_jaxConnector_imports_jax_if_installed():
    import piquasso as pq

    pq.JaxConnector()

    assert "jax" in sys.modules


def test_jaxConnector_raises_ImportError_if_jax_not_installed(
    raise_ImportError_when_importing_jax,
):
    import piquasso as pq

    with pytest.raises(ImportError) as error:
        pq.JaxConnector()

    assert error.value.args[0] == (
        "You have invoked a feature which requires 'jax'.\n"
        "You can install JAX via:\n"
        "\n"
        "pip install piquasso[jax]"
    )


def test_importing_Piquasso_does_not_import_jax(unimport_jax):
    import piquasso as pq  # noqa: F401

    assert "jax" not in sys.modules


def test_Piquasso_works_without_jax(
    unimport_jax,
    raise_ImportError_when_importing_jax,
):
    import piquasso as pq  # noqa: F401

    assert "jax" not in sys.modules
